"""
InfraSyncExecutor class which runs build, package and deploy contexts
"""
import copy
import logging
import re
from typing import Dict, Optional, Set

from boto3 import Session
from botocore.exceptions import ClientError

from samcli.commands._utils.template import get_template_data
from samcli.commands.build.build_context import BuildContext
from samcli.commands.deploy.deploy_context import DeployContext
from samcli.commands.package.package_context import PackageContext
from samcli.lib.providers.provider import ResourceIdentifier
from samcli.lib.providers.sam_stack_provider import is_local_path
from samcli.lib.utils.boto_utils import get_boto_client_provider_from_session_with_config
from samcli.lib.utils.resources import (
    AWS_APIGATEWAY_RESTAPI,
    AWS_APIGATEWAY_V2_API,
    AWS_CLOUDFORMATION_STACK,
    AWS_LAMBDA_FUNCTION,
    AWS_LAMBDA_LAYERVERSION,
    AWS_SERVERLESS_API,
    AWS_SERVERLESS_APPLICATION,
    AWS_SERVERLESS_FUNCTION,
    AWS_SERVERLESS_HTTPAPI,
    AWS_SERVERLESS_LAYERVERSION,
    AWS_SERVERLESS_STATEMACHINE,
    AWS_STEPFUNCTIONS_STATEMACHINE,
    CODE_SYNCABLE_RESOURCES,
    SYNCABLE_STACK_RESOURCES,
)
from samcli.yamlhelper import yaml_parse

LOG = logging.getLogger(__name__)

GENERAL_REMOVAL_MAP = {
    AWS_SERVERLESS_FUNCTION: ["CodeUri", "ImageUri"],
    AWS_SERVERLESS_LAYERVERSION: ["ContentUri"],
    AWS_LAMBDA_LAYERVERSION: ["Content"],
    AWS_SERVERLESS_API: ["DefinitionBody"],
    AWS_APIGATEWAY_RESTAPI: ["BodyS3Location"],
    AWS_SERVERLESS_HTTPAPI: ["DefinitionUri"],
    AWS_APIGATEWAY_V2_API: ["BodyS3Location"],
    AWS_SERVERLESS_STATEMACHINE: ["DefinitionUri"],
    AWS_STEPFUNCTIONS_STATEMACHINE: ["DefinitionS3Location"],
    AWS_SERVERLESS_APPLICATION: ["Location"],
    AWS_CLOUDFORMATION_STACK: ["TemplateURL"],
}

LAMBDA_FUNCTION_REMOVAL_MAP = {
    AWS_LAMBDA_FUNCTION: {"Code": ["ImageUri", "S3Bucket", "S3Key", "S3ObjectVersion"]},
}


class InfraSyncExecutor:
    """
    Executor for infra sync that contains skip logic when template is not changed
    """

    _build_context: BuildContext
    _package_context: PackageContext
    _deploy_context: DeployContext
    _code_sync_resources: Set[ResourceIdentifier]

    def __init__(self, build_context: BuildContext, package_context: PackageContext, deploy_context: DeployContext):
        """Constructs the sync for infra executor.

        Parameters
        ----------
        build_context : BuildContext
        package_context : PackageContext
        deploy_context : DeployContext
        """
        self._build_context = build_context
        self._package_context = package_context
        self._deploy_context = deploy_context

        self._code_sync_resources = set()

        session = Session(profile_name=self._deploy_context.profile, region_name=self._deploy_context.region)
        self._cfn_client = self._boto_client("cloudformation", session)
        self._s3_client = self._boto_client("s3", session)

    def _boto_client(self, client_name: str, session: Session):
        """
        Creates boto client

        Parameters
        ----------
        client_name: str
            The name of the client
        session: boto3.Session
            The session created using customer config

        Returns
        -------
        Service client instance
        """
        return get_boto_client_provider_from_session_with_config(session)(client_name)

    def _auto_skip_infra_sync(
        self,
        packaged_template_path: str,
        built_template_path: str,
        stack_name: str,
        nested_prefix: Optional[str] = None,
    ) -> bool:
        """
        Recursively compares two templates, including the nested templates referenced inside

        Parameters
        ----------
        packaged_template_path : str
            The template location of the current template packaged
        built_template_path : str
            The template location of the current template built
        stack_name : str
            The CloudFormation stack name that the template is deployed to
        nested_prefix: Optional[str]
            The nested stack stack name tree for child stack resources

        Returns
        -------
        bool
            Returns True if no template changes from last deployment
            Returns False if there are template differences
        """
        current_template = self.get_template(packaged_template_path)
        current_built_template = self.get_template(built_template_path)

        if not current_template or not current_built_template:
            LOG.debug("Cannot obtain a working current template for template path")
            return False

        try:
            last_deployed_template_str = self._cfn_client.get_template(
                StackName=stack_name, TemplateStage="Original"
            ).get("TemplateBody", "")
        except ClientError as ex:
            LOG.debug("Stack with name %s does not exist on CloudFormation", stack_name, exc_info=ex)
            return False

        last_deployed_template_dict = yaml_parse(last_deployed_template_str)

        sanitized_current_template = copy.deepcopy(current_template)
        sanitized_last_template = copy.deepcopy(last_deployed_template_dict)

        sanitized_resources = self._sanitize_template(
            sanitized_current_template, built_template_dict=current_built_template
        )
        self._sanitize_template(sanitized_last_template, linked_resources=sanitized_resources)

        if sanitized_last_template != sanitized_current_template:
            LOG.debug("The current template is different from the last deployed version, we will not skip infra sync")
            return False

        # The recursive template check for Nested stacks
        for resource_logical_id in current_template.get("Resources", {}):
            resource_dict = current_template.get("Resources", {}).get(resource_logical_id, {})
            resource_type = resource_dict.get("Type")

            if resource_type in CODE_SYNCABLE_RESOURCES:
                last_resource_dict = last_deployed_template_dict.get("Resources", {}).get(resource_logical_id, {})
                resource_resolved_id = nested_prefix + resource_logical_id if nested_prefix else resource_logical_id

                if resource_type == AWS_LAMBDA_FUNCTION:
                    if not resource_dict.get("Properties", {}).get("Code", None) == last_resource_dict.get(
                        "Properties", {}
                    ).get("Code", None):
                        self._code_sync_resources.add(ResourceIdentifier(resource_resolved_id))
                else:
                    for field in GENERAL_REMOVAL_MAP.get(resource_type, []):
                        if not resource_dict.get("Properties", {}).get(field, None) == last_resource_dict.get(
                            "Properties", {}
                        ).get(field, None):
                            self._code_sync_resources.add(ResourceIdentifier(resource_resolved_id))

            if resource_type in SYNCABLE_STACK_RESOURCES:
                try:
                    stack_resource_detail = self._cfn_client.describe_stack_resource(
                        StackName=stack_name, LogicalResourceId=resource_logical_id
                    )
                except ClientError as ex:
                    LOG.debug(
                        "Cannot get resource detail with name %s on CloudFormation", resource_logical_id, exc_info=ex
                    )
                    return False

                # If the nested stack is of type AWS::CloudFormation::Stack,
                # The template location will be under TemplateURL property
                # If the nested stack is of type AWS::Serverless::Application,
                # the template location will be under Location property
                template_field = "TemplateURL" if resource_type == AWS_CLOUDFORMATION_STACK else "Location"
                template_location = resource_dict.get("Properties", {}).get(template_field)

                # For AWS::Serverless::Application, location can be a ApplicationLocationObject dict containing SAR ID
                if isinstance(template_location, dict):
                    continue
                # For other scenarios, template location will be a string (local or s3 URL)
                elif not self._auto_skip_infra_sync(
                    resource_dict.get("Properties", {}).get(template_field),
                    current_built_template.get("Resources", {})
                    .get(resource_logical_id, {})
                    .get("Properties", {})
                    .get(template_field),
                    stack_resource_detail.get("StackResourceDetail", {}).get("PhysicalResourceId", ""),
                    nested_prefix + resource_logical_id + "/" if nested_prefix else resource_logical_id + "/",
                ):
                    return False

        LOG.debug("There are no changes from the previously deployed template for %s", packaged_template_path)
        return True

    def _sanitize_template(
        self, template_dict: Dict, linked_resources: Set[str] = set(), built_template_dict: Optional[Dict] = None
    ) -> Set[str]:
        """
        Fields skipped during template comparison because sync --code can handle the difference:
        * CodeUri or ImageUri property of AWS::Serverless::Function
        * ImageUri, S3Bucket, S3Key, S3ObjectVersion fields in Code property of AWS::Lambda::Function
        * ContentUri property of AWS::Serverless::LayerVersion
        * Content property of AWS::Lambda::LayerVersion
        * DefinitionBody property of AWS::Serverless::Api
        * BodyS3Location property of AWS::ApiGateway::RestApi
        * DefinitionUri property of AWS::Serverless::HttpApi
        * BodyS3Location property of AWS::ApiGatewayV2::Api
        * DefinitionUri property of AWS::Serverless::StateMachine
        * DefinitionS3Location property of AWS::StepFunctions::StateMachine

        Fields skipped during template comparison because we have recursive compare logic for nested stack:
        * Location property of AWS::Serverless::Application
        * TemplateURL property of AWS::CloudFormation::Stack

        Fields skipped during template comparison because it's a metadata generated by SAM
        * SamResourceId in Metadata property of all resources

        Parameters
        ----------
        template_dict: Dict
            The unprocessed template dictionary
        linked_resources: List[str]
            The corresponding resources in the other template that got processed
        built_template_dict: Optional[Dict]
            The built template dict that the paths didn't get replaced with packaged links yet

        Returns
        -------
        Set[str]
            The list of resource IDs that got changed during sanitization
        """

        resources = template_dict.get("Resources", {})
        processed_resources: Set[str] = set()

        for resource_logical_id in resources:
            resource_dict = resources.get(resource_logical_id, {})

            # Built resource dict helps with determining if a field is a local path
            if built_template_dict:
                built_resource_dict = built_template_dict.get("Resources", {}).get(resource_logical_id, {})

            resource_type = resource_dict.get("Type")

            if resource_type in CODE_SYNCABLE_RESOURCES or resource_type in SYNCABLE_STACK_RESOURCES:
                processed_resource = self._remove_resource_field(
                    resource_logical_id,
                    resource_type,
                    resource_dict,
                    linked_resources,
                    built_resource_dict if built_template_dict else None,
                )

                if processed_resource:
                    LOG.debug("Sanitized %s resource %s", resource_type, resource_logical_id)
                    processed_resources.add(processed_resource)

            # Remove SamResourceId metadata since this metadata does not affect any cloud behaviour
            resource_dict.get("Metadata", {}).pop("SamResourceId", None)
            if not resource_dict.get("Metadata"):
                resource_dict.pop("Metadata", None)
            LOG.debug("Sanitizing the Metadata for resource %s", resource_logical_id)

        return processed_resources

    def _remove_resource_field(
        self,
        resource_logical_id: str,
        resource_type: str,
        resource_dict: Dict,
        linked_resources: Set[str] = set(),
        built_resource_dict: Optional[Dict] = None,
    ) -> Optional[str]:
        """
        Helper method to process resource dict

        Parameters
        ----------
        resource_logical_id: str
            Logical ID of the resource
        resource_type: str
            Resource type
        resource_dict: Dict
            The resource level dict containing Properties field
        linked_resources: Set[str]
            The corresponding resources in the other template that got processed
        built_resource_dict: Optional[Dict]
            Only passed in for current template sanitization to determine if local

        Returns
        -------
        Optional[str]
            The processed resource ID
        """
        processed_logical_id = None

        if resource_type == AWS_LAMBDA_FUNCTION:
            for field in LAMBDA_FUNCTION_REMOVAL_MAP.get(resource_type, {}).get("Code", []):
                # We sanitize only if the provided resource is local
                if (
                    built_resource_dict
                    and is_local_path(built_resource_dict.get("Properties", {}).get("Code", {}).get(field, None))
                ) or resource_logical_id in linked_resources:
                    resource_dict.get("Properties", {}).get("Code", {}).pop(field, None)
                    processed_logical_id = resource_logical_id
        else:
            for field in GENERAL_REMOVAL_MAP.get(resource_type, []):
                if resource_type in SYNCABLE_STACK_RESOURCES:
                    if not isinstance(resource_dict.get("Properties", {}).get(field, None), dict):
                        resource_dict.get("Properties", {}).pop(field, None)
                        processed_logical_id = resource_logical_id
                elif (
                    built_resource_dict and is_local_path(built_resource_dict.get("Properties", {}).get(field, None))
                ) or resource_logical_id in linked_resources:
                    resource_dict.get("Properties", {}).pop(field, None)
                    processed_logical_id = resource_logical_id

        return processed_logical_id

    def get_template(self, template_path: str) -> Optional[Dict]:
        """
        Returns the template dict based on local or remote read logic

        Parameters
        ----------
        template_path: str
            The location of the template

        Returns
        -------
        Dict
            The parsed template dict
        """
        template = None
        # If the customer template uses a nested stack with location/template URL in S3
        if template_path.startswith("https://"):
            template = self._get_remote_template_data(template_path)

        # If the template location is local
        else:
            template = get_template_data(template_path)

        return template

    def _get_remote_template_data(self, template_path: str) -> Optional[Dict]:
        """
        Get template dict from remote location

        Parameters
        ----------
        template_path: str
            The s3 location of the template

        Returns
        -------
        Dict
            The parsed template dict from s3
        """
        template = None

        parsed_s3_location = re.search(r"https:\/\/[^/]*\/([^/]*)\/(.*)", template_path)
        if parsed_s3_location:
            s3_bucket = parsed_s3_location.group(1)
            s3_key = parsed_s3_location.group(2)
            try:
                s3_object = self._s3_client.get_object(Bucket=s3_bucket, Key=s3_key)
            except ClientError as ex:
                LOG.debug("The provided template location %s can not be found", template_path, exc_info=ex)
            else:
                streaming_body = s3_object.get("Body")
                if streaming_body:
                    template = yaml_parse(streaming_body.read().decode("utf-8"))

        return template

    @property
    def code_sync_resources(self) -> Set[ResourceIdentifier]:
        """Returns the list of resources that should trigger code sync"""
        return self._code_sync_resources



"""
Represents a pipeline OIDC provider
"""
from abc import abstractmethod
from typing import List

import click

from samcli.commands.pipeline.bootstrap.guided_context import BITBUCKET, GITHUB_ACTIONS, GITLAB
from samcli.lib.config.samconfig import SamConfig


class PipelineOidcProvider:

    PROVIDER_URL_PARAMETER = "oidc-provider-url"
    CLIENT_ID_PARAMETER = "oidc-client-id"
    OPENID_CONNECT = "OpenID Connect (OIDC)"

    def __init__(self, oidc_parameters: dict, oidc_parameter_names: List[str], oidc_provider_name: str) -> None:
        self.oidc_parameters = oidc_parameters
        self.oidc_parameter_names = [self.PROVIDER_URL_PARAMETER, self.CLIENT_ID_PARAMETER] + oidc_parameter_names
        self.oidc_provider_name = oidc_provider_name
        self.verify_parameters()

    def verify_parameters(self) -> None:
        """
        Makes sure that all required parameters have been provided
        -------
        """
        error_string = ""
        for parameter_name in self.oidc_parameter_names:
            if not self.oidc_parameters[parameter_name]:
                error_string += f"Missing required parameter '--{parameter_name}'\n"
        if error_string:
            raise click.UsageError("\n" + error_string)

    def save_values(self, samconfig: SamConfig, cmd_names: List[str], section: str) -> None:
        """
        Saves provided values into config file so they can be reused for future calls to bootstrap
        """
        for parameter_name in self.oidc_parameter_names:
            samconfig.put(
                cmd_names=cmd_names,
                section=section,
                key=parameter_name.replace("-", "_"),
                value=self.oidc_parameters[parameter_name],
            )
        samconfig.put(cmd_names=cmd_names, section=section, key="oidc_provider", value=self.oidc_provider_name)
        samconfig.put(cmd_names=cmd_names, section=section, key="permissions_provider", value=self.OPENID_CONNECT)

    @abstractmethod
    def get_subject_claim(self) -> str:
        pass


class GitHubOidcProvider(PipelineOidcProvider):
    """
    Represents a GitHub Actions OIDC provider
    https://docs.github.com/en/actions/deployment/security-hardening-your-deployments/about-security-hardening-with-openid-connect
    ----------
    subject_claim_parameters: dict
        Parameters specific to building the subject claim for this provider.
    oidc_parameters: dict
        Parameters common to all providers.
    """

    GITHUB_ORG_PARAMETER_NAME = "github-org"
    GITHUB_REPO_PARAMETER_NAME = "github-repo"
    DEPLOYMENT_BRANCH_PARAMETER_NAME = "deployment-branch"

    def __init__(self, subject_claim_parameters: dict, oidc_parameters: dict) -> None:
        all_oidc_parameters = {**oidc_parameters, **subject_claim_parameters}
        all_oidc_parameter_names = [
            self.GITHUB_ORG_PARAMETER_NAME,
            self.GITHUB_REPO_PARAMETER_NAME,
            self.DEPLOYMENT_BRANCH_PARAMETER_NAME,
        ]
        super().__init__(all_oidc_parameters, all_oidc_parameter_names, GITHUB_ACTIONS)

    def get_subject_claim(self) -> str:
        """
        Returns the subject claim that will be used to establish trust between the OIDC provider and AWS.
        To read more about OIDC claims see the following: https://openid.net/specs/openid-connect-core-1_0.html#Claims
        https://tinyurl.com/github-oidc-token
        In GitHubs case when using the official OIDC action to assume a role the audience claim will always be
        sts.amazon.aws so we must use the subject claim https://tinyurl.com/github-oidc-claim
        -------
        """
        org = self.oidc_parameters["github-org"]
        repo = self.oidc_parameters["github-repo"]
        branch = self.oidc_parameters["deployment-branch"]
        return f"repo:{org}/{repo}:ref:refs/heads/{branch}"


class GitLabOidcProvider(PipelineOidcProvider):
    """
    Represents a GitLab OIDC provider
    https://docs.gitlab.com/ee/integration/openid_connect_provider.html
    ----------
    subject_claim_parameters: dict
        Parameters specific to building the subject claim for this provider.
    oidc_parameters: dict
        Parameters common to all providers.
    """

    GITLAB_PROJECT_PARAMETER_NAME = "gitlab-project"
    GITLAB_GROUP_PARAMETER_NAME = "gitlab-group"
    DEPLOYMENT_BRANCH_PARAMETER_NAME = "deployment-branch"

    def __init__(self, subject_claim_parameters: dict, oidc_parameters: dict) -> None:
        all_oidc_parameters = {**oidc_parameters, **subject_claim_parameters}
        all_oidc_parameter_names = [
            self.GITLAB_PROJECT_PARAMETER_NAME,
            self.GITLAB_GROUP_PARAMETER_NAME,
            self.DEPLOYMENT_BRANCH_PARAMETER_NAME,
        ]
        super().__init__(all_oidc_parameters, all_oidc_parameter_names, GITLAB)

    def get_subject_claim(self) -> str:
        """
        Returns the subject claim that will be used to establish trust between the OIDC provider and AWS.
        To read more about OIDC claims see the following: https://openid.net/specs/openid-connect-core-1_0.html#Claims
        https://docs.gitlab.com/ee/ci/cloud_services/aws/#configure-a-role-and-trust
        To learn more about configuring a role to work with GitLab OIDC through claims see the following
        https://docs.gitlab.com/ee/ci/cloud_services/index.html#configure-a-conditional-role-with-oidc-claims
        -------
        """
        group = self.oidc_parameters["gitlab-group"]
        project = self.oidc_parameters["gitlab-project"]
        branch = self.oidc_parameters["deployment-branch"]
        return f"project_path:{group}/{project}:ref_type:branch:ref:{branch}"


class BitbucketOidcProvider(PipelineOidcProvider):
    """
    Represents a Bitbucket OIDC provider
    https://support.atlassian.com/bitbucket-cloud/docs/integrate-pipelines-with-resource-servers-using-oidc/
    ----------
    subject_claim_parameters: dict
        Parameters specific to building the subject claim for this provider.
    oidc_parameters: dict
        Parameters common to all providers.
    """

    BITBUCKET_REPO_UUID_PARAMETER_NAME = "bitbucket-repo-uuid"

    def __init__(self, subject_claim_parameters: dict, oidc_parameters: dict) -> None:
        all_oidc_parameters = {**oidc_parameters, **subject_claim_parameters}
        all_oidc_parameter_names = [
            self.BITBUCKET_REPO_UUID_PARAMETER_NAME,
        ]
        super().__init__(all_oidc_parameters, all_oidc_parameter_names, BITBUCKET)

    def get_subject_claim(self) -> str:
        """
        Returns the subject claim that will be used to establish trust between the OIDC provider and AWS.
        To read more about OIDC claims see the following: https://openid.net/specs/openid-connect-core-1_0.html#Claims
        To learn more about configuring a role to work with GitLab OIDC through claims see the following
        tinyurl.com/bitbucket-oidc-claims
        -------
        """
        repo_uuid = self.oidc_parameters[self.BITBUCKET_REPO_UUID_PARAMETER_NAME]
        return f"{repo_uuid}:*"



""" Represents AWS resource"""
from typing import Optional


class ARNParts:
    """
    Decompose a given ARN into its parts https://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html

    Attributes
    ----------
    partition: str
        the partition part(AWS, aws-cn or aws-us-gov)  of the ARN
    service: str
        the service part(S3, IAM, ECR, ...etc) of the ARN
    region: str
        the AWS region part(us-east-1, eu-west-1, ...etc) of the ARN
    account-id: str
        the account-id part of the ARN
    resource-id: str
        the resource-id part of the ARN
    resource-type: str
        the resource-type part of the ARN
    """

    partition: str
    service: str
    region: str
    account_id: str
    resource_id: str

    def __init__(self, arn: str) -> None:
        parts = arn.split(":")
        try:
            [_, self.partition, self.service, self.region, self.account_id, self.resource_id] = parts
        except ValueError as ex:
            raise ValueError(f"Invalid ARN ({arn})") from ex


class Resource:
    """
    Represents an AWS resource

    Attributes
    ----------
    arn: str
        the ARN of the resource
    comment: str
        the comment of the resource
    is_user_provided: bool
        True if the user provided the ARN of the resource during the initialization. It indicates whether this pipeline-
        resource is provided by the user or created by SAM during `sam pipeline bootstrap`

    Methods
    -------
    name(self) -> Optional[str]:
        extracts and returns the resource name from its ARN
    """

    def __init__(self, arn: Optional[str], comment: Optional[str]) -> None:
        self.arn: Optional[str] = arn
        self.comment: Optional[str] = comment
        self.is_user_provided: bool = bool(arn)

    def name(self) -> Optional[str]:
        """
        extracts and returns the resource name from its ARN
        Raises
        ------
        ValueError if the ARN is invalid
        """
        if not self.arn:
            return None
        arn_parts: ARNParts = ARNParts(arn=self.arn)
        return arn_parts.resource_id


class IAMUser(Resource):
    """
    Represents an AWS IAM User resource
    Attributes
    ----------
    access_key_id: Optional[str]
        holds the AccessKeyId of the credential of this IAM user, if any.
    secret_access_key: Optional[str]
        holds the SecretAccessKey of the credential of this IAM user, if any.
    """

    def __init__(
        self,
        arn: Optional[str],
        comment: Optional[str],
        access_key_id: Optional[str] = None,
        secret_access_key: Optional[str] = None,
    ) -> None:
        self.access_key_id: Optional[str] = access_key_id
        self.secret_access_key: Optional[str] = secret_access_key
        super().__init__(arn=arn, comment=comment)


class S3Bucket(Resource):
    """
    Represents an AWS S3Bucket resource
    Attributes
    ----------
    kms_key_arn: Optional[str]
        The ARN of the KMS key used in encrypting this S3Bucket, if any.
    """

    def __init__(self, arn: Optional[str], comment: Optional[str], kms_key_arn: Optional[str] = None) -> None:
        self.kms_key_arn: Optional[str] = kms_key_arn
        super().__init__(arn=arn, comment=comment)


class ECRImageRepository(Resource):
    """Represents an AWS ECR image repository resource"""

    def __init__(self, arn: Optional[str], comment: Optional[str]) -> None:
        super().__init__(arn=arn, comment=comment)

    def get_uri(self) -> Optional[str]:
        """
        extracts and returns the URI of the given ECR image repository from its ARN
        see https://docs.aws.amazon.com/AmazonECR/latest/userguide/Registries.html
        Raises
        ------
        ValueError if the ARN is invalid
        """
        if not self.arn:
            return None
        arn_parts: ARNParts = ARNParts(self.arn)
        # ECR's resource_id contains the resource-type("resource") which is excluded from the URL
        # from docs: https://docs.aws.amazon.com/AmazonECR/latest/userguide/security_iam_service-with-iam.html
        # ECR's ARN: arn:${Partition}:ecr:${Region}:${Account}:repository/${Repository-name}
        if not arn_parts.resource_id.startswith("repository/"):
            raise ValueError(f"Invalid ECR ARN ({self.arn}), can't extract the URL from it.")
        i = len("repository/")
        repo_name = arn_parts.resource_id[i:]
        return f"{arn_parts.account_id}.dkr.ecr.{arn_parts.region}.amazonaws.com/{repo_name}"


class OidcProvider(Resource):
    """
    Represents an AWS OIDC Provider resource
    Attributes
    ----------
    client_id: str
        the client id used to authenticate the user with the OIDC provider.
    provider_url: str
        url of the OIDC provider.
    thumbprint: str
        thumbprint for the top intermediate certificate authority (CA)
         that signed the certificate used by the identity provider
    """

    def __init__(
        self,
        arn: Optional[str],
        comment: Optional[str],
        client_id: Optional[str],
        provider_url: Optional[str],
        thumbprint: Optional[str],
    ) -> None:
        self.client_id: Optional[str] = client_id
        self.provider_url: Optional[str] = provider_url
        self.thumbprint: Optional[str] = thumbprint
        super().__init__(arn=arn, comment=comment)



"""DefinitionValidator for Validating YAML and JSON Files"""
import logging
from pathlib import Path
from typing import Any, Dict, Optional

import yaml

from samcli.yamlhelper import parse_yaml_file

LOG = logging.getLogger(__name__)


class DefinitionValidator:
    _path: Path
    _detect_change: bool
    _data: Optional[Dict[str, Any]]

    def __init__(self, path: Path, detect_change: bool = True, initialize_data: bool = True) -> None:
        """
        Validator for JSON and YAML files.
        Calling validate_change() will return True if the definition is valid and has changes.

        Parameters
        ----------
        path : Path
            Path to the definition file
        detect_change : bool, optional
            validation will only be successful if there are changes between current and previous data,
            by default True
        initialize_data : bool, optional
            Should initialize existing definition data before the first validate, by default True
            Used along with detect_change
        """
        super().__init__()
        self._path = path
        self._detect_change = detect_change
        self._data = None
        if initialize_data:
            self.validate_change()

    def validate_change(self) -> bool:
        """Validate change on json or yaml file.

        Returns
        -------
        bool
            True if it is valid, False otherwise.
            If detect_change is set, False will also be returned if there is
            no change compared to the previous validation.
        """
        old_data = self._data

        if not self.validate_file():
            return False
        return old_data != self._data if self._detect_change else True

    def validate_file(self) -> bool:
        """Validate json or yaml file.

        Returns
        -------
        bool
            True if it is valid path and yaml file, False otherwise.
        """
        if not self._path.exists():
            LOG.debug(
                "File %s failed to validate due to file path does not exist. Please verify that the path is valid.",
                self._path,
            )
            return False

        try:
            self._data = parse_yaml_file(str(self._path))
        except (ValueError, yaml.YAMLError) as e:
            LOG.debug(
                "File %s failed to validate due to it file cannot be parsed. \
Please verify that file is in the correct json or yaml format.",
                self._path,
                exc_info=e,
            )
            return False
        return True



"""
Represents Lambda runtime containers.
"""
import logging
from typing import List

from samcli.lib.utils.packagetype import IMAGE
from samcli.local.docker.lambda_debug_settings import LambdaDebugSettings

from .container import Container
from .lambda_image import LambdaImage, Runtime

LOG = logging.getLogger(__name__)


class LambdaContainer(Container):
    """
    Represents a Lambda runtime container. This class knows how to setup entry points, environment variables,
    exposed ports etc specific to Lambda runtime container. The container management functionality (create/start/stop)
    is provided by the base class
    """

    _WORKING_DIR = "/var/task"
    _DEFAULT_ENTRYPOINT = ["/var/rapid/aws-lambda-rie", "--log-level", "error"]

    # The Volume Mount path for debug files in docker
    _DEBUGGER_VOLUME_MOUNT_PATH = "/tmp/lambci_debug_files"
    _DEFAULT_CONTAINER_DBG_GO_PATH = _DEBUGGER_VOLUME_MOUNT_PATH + "/dlv"

    # Options for selecting debug entry point
    _DEBUG_ENTRYPOINT_OPTIONS = {"delvePath": _DEFAULT_CONTAINER_DBG_GO_PATH}

    # This is the dictionary that represents where the debugger_path arg is mounted in docker to as readonly.
    _DEBUGGER_VOLUME_MOUNT = {"bind": _DEBUGGER_VOLUME_MOUNT_PATH, "mode": "ro"}

    def __init__(
        self,  # pylint: disable=R0914
        runtime,
        imageuri,
        handler,
        packagetype,
        image_config,
        code_dir,
        layers,
        lambda_image,
        architecture,
        memory_mb=128,
        env_vars=None,
        debug_options=None,
        container_host=None,
        container_host_interface=None,
        function_full_path=None,
    ):
        """
        Initializes the class

        Parameters
        ----------
        runtime str
            Name of the Lambda runtime
        imageuri str
            Name of the Lambda Image which is of the form {image}:{tag}
        handler str
            Handler of the function to run
        packagetype str
            Package type for the lambda function which is either zip or image.
        image_config dict
            Image configuration which can be used set to entrypoint, command and working dir for the container.
        code_dir str
            Directory where the Lambda function code is present. This directory will be mounted
            to the container to execute
        layers list(str)
            List of layers
        lambda_image samcli.local.docker.lambda_image.LambdaImage
            LambdaImage that can be used to build the image needed for starting the container
        architecture str
            Architecture type either x86_64 or arm64 on AWS lambda
        memory_mb int
            Optional. Max limit of memory in MegaBytes this Lambda function can use.
        env_vars dict
            Optional. Dictionary containing environment variables passed to container
        debug_options DebugContext
            Optional. Contains container debugging info (port, debugger path)
        container_host string
            Optional. Host of locally emulated Lambda container
        container_host_interface
            Optional. Interface that Docker host binds ports to
        function_full_path str
            Optional. The function full path, unique in all stacks
        """
        if not Runtime.has_value(runtime) and not packagetype == IMAGE:
            raise ValueError("Unsupported Lambda runtime {}".format(runtime))

        image = LambdaContainer._get_image(
            lambda_image, runtime, packagetype, imageuri, layers, architecture, function_full_path
        )
        ports = LambdaContainer._get_exposed_ports(debug_options)
        config = LambdaContainer._get_config(lambda_image, image)
        entry, container_env_vars = LambdaContainer._get_debug_settings(runtime, debug_options)
        additional_options = LambdaContainer._get_additional_options(runtime, debug_options)
        additional_volumes = LambdaContainer._get_additional_volumes(runtime, debug_options)

        _work_dir = self._WORKING_DIR
        _entrypoint = None
        _command = None
        if not env_vars:
            env_vars = {}

        if packagetype == IMAGE:
            _command = (image_config.get("Command") if image_config else None) or config.get("Cmd")
            if not env_vars.get("AWS_LAMBDA_FUNCTION_HANDLER", None):
                # NOTE(sriram-mv):
                # Set AWS_LAMBDA_FUNCTION_HANDLER to be based of the command for Image based Packagetypes.
                env_vars["AWS_LAMBDA_FUNCTION_HANDLER"] = _command[0] if isinstance(_command, list) else None
            _additional_entrypoint_args = (image_config.get("EntryPoint") if image_config else None) or config.get(
                "Entrypoint"
            )
            _entrypoint = entry or self._DEFAULT_ENTRYPOINT
            # NOTE(sriram-mv): Only add entrypoint specified in the image configuration if the entrypoint
            # has not changed for debugging.
            if isinstance(_additional_entrypoint_args, list) and entry == self._DEFAULT_ENTRYPOINT:
                _entrypoint = _entrypoint + _additional_entrypoint_args
            _work_dir = (image_config.get("WorkingDirectory") if image_config else None) or config.get("WorkingDir")

        env_vars = {**env_vars, **container_env_vars}
        super().__init__(
            image,
            _command if _command else [],
            _work_dir,
            code_dir,
            memory_limit_mb=memory_mb,
            exposed_ports=ports,
            entrypoint=_entrypoint if _entrypoint else entry,
            env_vars=env_vars,
            container_opts=additional_options,
            additional_volumes=additional_volumes,
            container_host=container_host,
            container_host_interface=container_host_interface,
        )

    @staticmethod
    def _get_exposed_ports(debug_options):
        """
        Return Docker container port binding information. If a debug port tuple is given, then we will ask Docker to
        bind every given port to same port both inside and outside the container ie.
        Runtime process is started in debug mode with at given port inside the container
        and exposed to the host machine at the same port.

        :param DebugContext debug_options: Debugging options for the function (includes debug port, args, and path)
        :return dict: Dictionary containing port binding information. None, if debug_port was not given
        """
        if not debug_options:
            return None

        if not debug_options.debug_ports:
            return None

        # container port : host port
        ports_map = {}
        for port in debug_options.debug_ports:
            ports_map[port] = port

        return ports_map

    @staticmethod
    def _get_additional_options(runtime: str, debug_options):
        """
        Return additional Docker container options. Used by container debug mode to enable certain container
        security options.
        :param runtime: The runtime string
        :param DebugContext debug_options: DebugContext for the runtime of the container.
        :return dict: Dictionary containing additional arguments to be passed to container creation.
        """
        if not debug_options:
            return None

        opts = {}

        if runtime == Runtime.go1x.value:
            # These options are required for delve to function properly inside a docker container on docker < 1.12
            # See https://github.com/moby/moby/issues/21051
            opts["security_opt"] = ["seccomp:unconfined"]
            opts["cap_add"] = ["SYS_PTRACE"]

        return opts

    @staticmethod
    def _get_additional_volumes(runtime, debug_options):
        """
        Return additional volumes to be mounted in the Docker container. Used by container debug for mapping
        debugger executable into the container.
        :param runtime: the runtime string
        :param DebugContext debug_options: DebugContext for the runtime of the container.
        :return dict: Dictionary containing volume map passed to container creation.
        """
        volumes = {}

        if debug_options and debug_options.debugger_path:
            volumes[debug_options.debugger_path] = LambdaContainer._DEBUGGER_VOLUME_MOUNT

        return volumes

    @staticmethod
    def _get_image(
        lambda_image: LambdaImage,
        runtime: str,
        packagetype: str,
        image: str,
        layers: List[str],
        architecture: str,
        function_name: str,
    ):
        """

        Parameters
        ----------
        lambda_image : LambdaImage
            LambdaImage that can be used to build the image needed for starting the container
        runtime : str
            Name of the Lambda runtime
        packagetype : str
            Package type for the lambda function which is either zip or image.
        image : str
            Name of the Lambda Image which is of the form {image}:{tag}
        layers : List[str]
            List of layers
        architecture : str
            Architecture type either x86_64 or arm64 on AWS lambda
        function_name: str
            The name of the lambda function that the container is to invoke

        Returns
        -------
        str
            Name of Docker Image for the given runtime
        """
        return lambda_image.build(runtime, packagetype, image, layers, architecture, function_name=function_name)

    @staticmethod
    def _get_config(lambda_image, image):
        return lambda_image.get_config(image)

    @staticmethod
    def _get_debug_settings(runtime, debug_options=None):  # pylint: disable=too-many-branches
        """
        Returns the entry point for the container. The default value for the entry point is already configured in the
        Dockerfile. We override this default specifically when enabling debugging. The overridden entry point includes
        a few extra flags to start the runtime in debug mode.

        :param string runtime: Lambda function runtime name.
        :param DebugContext debug_options: Optional. Debug context for the function (includes port, args, and path).
        :return list: List containing the new entry points. Each element in the list is one portion of the command.
            ie. if command is ``node index.js arg1 arg2``, then this list will be ["node", "index.js", "arg1", "arg2"]
        """

        entry = LambdaContainer._DEFAULT_ENTRYPOINT
        if not debug_options:
            return entry, {}

        debug_ports = debug_options.debug_ports
        container_env_vars = debug_options.container_env_vars
        if not debug_ports:
            return entry, {}

        debug_port = debug_ports[0]
        debug_args_list = []

        if debug_options.debug_args:
            debug_args_list = debug_options.debug_args.split(" ")
        # configs from: https://github.com/lambci/docker-lambda
        # to which we add the extra debug mode options
        return LambdaDebugSettings.get_debug_settings(
            debug_port=debug_port,
            debug_args_list=debug_args_list,
            _container_env_vars=container_env_vars,
            runtime=runtime,
            options=LambdaContainer._DEBUG_ENTRYPOINT_OPTIONS,
        )



"""Handles Swagger Parsing"""

import logging

from samcli.commands.local.lib.swagger.integration_uri import IntegrationType, LambdaUri
from samcli.local.apigw.local_apigw_service import Route

LOG = logging.getLogger(__name__)


class SwaggerParser:
    _INTEGRATION_KEY = "x-amazon-apigateway-integration"
    _ANY_METHOD_EXTENSION_KEY = "x-amazon-apigateway-any-method"
    _BINARY_MEDIA_TYPES_EXTENSION_KEY = "x-amazon-apigateway-binary-media-types"  # pylint: disable=C0103
    _ANY_METHOD = "ANY"

    def __init__(self, stack_path: str, swagger):
        """
        Constructs an Swagger Parser object

        :param str stack_path: Path of the stack the resource is located
        :param dict swagger: Dictionary representation of a Swagger document
        """
        self.swagger = swagger or {}
        self.stack_path = stack_path

    def get_binary_media_types(self):
        """
        Get the list of Binary Media Types from Swagger

        Returns
        -------
        list of str
            List of strings that represent the Binary Media Types for the API, defaulting to empty list is None

        """
        return self.swagger.get(self._BINARY_MEDIA_TYPES_EXTENSION_KEY) or []

    def get_routes(self, event_type=Route.API):
        """
        Parses a swagger document and returns a list of APIs configured in the document.

        Swagger documents have the following structure
        {
            "/path1": {    # path
                "get": {   # method
                    "x-amazon-apigateway-integration": {   # integration
                        "type": "aws_proxy",

                        # URI contains the Lambda function ARN that needs to be parsed to get Function Name
                        "uri": {
                            "Fn::Sub":
                                "arn:aws:apigateway:aws:lambda:path/2015-03-31/functions/${LambdaFunction.Arn}/..."
                        }
                    }
                },
                "post": {
                },
            },
            "/path2": {
                ...
            }
        }

        Returns
        -------
        list of list of samcli.commands.local.apigw.local_apigw_service.Route
            List of APIs that are configured in the Swagger document
        """

        result = []
        paths_dict = self.swagger.get("paths", {})

        for full_path, path_config in paths_dict.items():
            for method, method_config in path_config.items():

                function_name = self._get_integration_function_name(method_config)
                if not function_name:
                    LOG.debug(
                        "Lambda function integration not found in Swagger document at path='%s' method='%s'",
                        full_path,
                        method,
                    )
                    continue

                if method.lower() == self._ANY_METHOD_EXTENSION_KEY:
                    # Convert to a more commonly used method notation
                    method = self._ANY_METHOD
                payload_format_version = self._get_payload_format_version(method_config)
                route = Route(
                    function_name,
                    full_path,
                    methods=[method],
                    event_type=event_type,
                    payload_format_version=payload_format_version,
                    operation_name=method_config.get("operationId"),
                    stack_path=self.stack_path,
                )
                result.append(route)
        return result

    def _get_integration(self, method_config):
        """
        Get Integration defined in the method configuration.
        Integration configuration is defined under the special "x-amazon-apigateway-integration" key. We care only
        about Lambda integrations, which are of type aws_proxy, and ignore the rest.

        Parameters
        ----------
        method_config : dict
            Dictionary containing the method configuration which might contain integration settings

        Returns
        -------
        dict or None
            integration, if possible. None, if not.
        """
        if not isinstance(method_config, dict) or self._INTEGRATION_KEY not in method_config:
            return None

        integration = method_config[self._INTEGRATION_KEY]

        if (
            integration
            and isinstance(integration, dict)
            and integration.get("type").lower() == IntegrationType.aws_proxy.value
        ):
            # Integration must be "aws_proxy" otherwise we don't care about it
            return integration

        return None

    def _get_integration_function_name(self, method_config):
        """
        Tries to parse the Lambda Function name from the Integration defined in the method configuration.
        Integration configuration is defined under the special "x-amazon-apigateway-integration" key. We care only
        about Lambda integrations, which are of type aws_proxy, and ignore the rest. Integration URI is complex and
        hard to parse. Hence we do our best to extract function name out of integration URI. If not possible, we
        return None.

        Parameters
        ----------
        method_config : dict
            Dictionary containing the method configuration which might contain integration settings

        Returns
        -------
        string or None
            Lambda function name, if possible. None, if not.
        """
        integration = self._get_integration(method_config)
        if integration is None:
            return None

        return LambdaUri.get_function_name(integration.get("uri"))

    def _get_payload_format_version(self, method_config):
        """
        Get the "payloadFormatVersion" from the Integration defined in the method configuration.

        Parameters
        ----------
        method_config : dict
            Dictionary containing the method configuration which might contain integration settings

        Returns
        -------
        string or None
            Payload format version, if exists. None, if not.
        """
        integration = self._get_integration(method_config)
        if integration is None:
            return None

        return integration.get("payloadFormatVersion")



"""
CLI command for "pipeline init" command
"""
from typing import Any, Optional

import click

from samcli.cli.cli_config_file import TomlProvider, configuration_option
from samcli.cli.main import common_options as cli_framework_options
from samcli.cli.main import pass_context
from samcli.commands._utils.command_exception_handler import command_exception_handler
from samcli.commands.pipeline.init.interactive_init_flow import InteractiveInitFlow
from samcli.lib.telemetry.metric import track_command

SHORT_HELP = "Generates a CI/CD pipeline configuration file."
HELP_TEXT = """
This command generates a pipeline configuration file that your CI/CD system can use to deploy
serverless applications using AWS SAM.

Before using sam pipeline init, you must bootstrap the necessary resources for each stage in your pipeline.
You can do this by running sam pipeline init --bootstrap to be guided through the setup and configuration
file generation process, or refer to resources you have previously created with the sam pipeline bootstrap command.
"""


@click.command("init", help=HELP_TEXT, short_help=SHORT_HELP)
@configuration_option(provider=TomlProvider(section="parameters"))
@click.option(
    "--bootstrap",
    is_flag=True,
    default=False,
    help="Enable interactive mode that walks the user through creating necessary AWS infrastructure resources.",
)
@cli_framework_options
@pass_context
@track_command  # pylint: disable=R0914
@command_exception_handler
def cli(ctx: Any, config_env: Optional[str], config_file: Optional[str], bootstrap: bool) -> None:
    """
    `sam pipeline init` command entry point
    """

    # Currently we support interactive mode only, i.e. the user doesn't provide the required arguments during the call
    # so we call do_cli without any arguments. This will change after supporting the non interactive mode.
    do_cli(bootstrap)


def do_cli(bootstrap: bool) -> None:
    """
    implementation of `sam pipeline init` command
    """
    # TODO non-interactive mode
    init_flow = InteractiveInitFlow(bootstrap)
    init_flow.do_interactive()



"""SyncFlow for RestApi"""
import logging
from typing import TYPE_CHECKING, Dict, List, Optional, Set, cast

from botocore.exceptions import ClientError

from samcli.lib.providers.exceptions import MissingLocalDefinition
from samcli.lib.providers.provider import ResourceIdentifier, Stack, get_resource_by_id, get_resource_ids_by_type
from samcli.lib.sync.flows.generic_api_sync_flow import GenericApiSyncFlow
from samcli.lib.utils.colors import Colored
from samcli.lib.utils.resources import AWS_APIGATEWAY_DEPLOYMENT, AWS_APIGATEWAY_STAGE, AWS_SERVERLESS_API

# BuildContext and DeployContext will only be imported for type checking to improve performance
# since no instances of contexts will be instantiated in this class
if TYPE_CHECKING:  # pragma: no cover
    from samcli.commands.build.build_context import BuildContext
    from samcli.commands.deploy.deploy_context import DeployContext
    from samcli.commands.sync.sync_context import SyncContext

LOG = logging.getLogger(__name__)


class RestApiSyncFlow(GenericApiSyncFlow):
    """SyncFlow for RestApi's"""

    def __init__(
        self,
        restapi_identifier: str,
        build_context: "BuildContext",
        deploy_context: "DeployContext",
        sync_context: "SyncContext",
        physical_id_mapping: Dict[str, str],
        stacks: List[Stack],
    ):
        """
        Parameters
        ----------
        restapi_identifier : str
            RestApi resource identifier that needs to have associated RestApi updated.
        build_context : BuildContext
            BuildContext used for build related parameters
        deploy_context : BuildContext
            DeployContext used for this deploy related parameters
        sync_context: SyncContext
            SyncContext object that obtains sync information.
        physical_id_mapping : Dict[str, str]
            Mapping between resource logical identifier and physical identifier
        stacks : List[Stack], optional
            List of stacks containing a root stack and optional nested stacks
        """
        super().__init__(
            restapi_identifier,
            build_context,
            deploy_context,
            sync_context,
            physical_id_mapping,
            log_name="RestApi " + restapi_identifier,
            stacks=stacks,
        )
        self._api_physical_id = ""

    def set_up(self) -> None:
        super().set_up()
        self._api_client = self._boto_client("apigateway")
        self._api_physical_id = self.get_physical_id(self._api_identifier)

    def sync(self) -> None:
        if self._definition_uri is None:
            raise MissingLocalDefinition(ResourceIdentifier(self._api_identifier), "DefinitionUri")

        self._update_api()
        new_dep_id = self._create_deployment()
        stages = self._collect_stages()
        prev_dep_ids = self._update_stages(stages, new_dep_id)
        self._delete_deployments(prev_dep_ids)

    def _update_api(self) -> None:
        """
        Update the API content
        """
        LOG.debug("%sTrying to update RestAPI through client", self.log_prefix)
        response_put = cast(
            Dict,
            self._api_client.put_rest_api(restApiId=self._api_physical_id, mode="overwrite", body=self._swagger_body),
        )
        LOG.debug("%sPut RestApi Result: %s", self.log_prefix, response_put)

    def _create_deployment(self) -> Optional[str]:
        """
        Create a deployment using the updated API and record the created deployment ID

        Returns
        ----------
        Optional[str]: The newly created deployment ID
        """
        LOG.debug("%sTrying to create a deployment through client", self.log_prefix)
        response_dep = cast(
            Dict, self._api_client.create_deployment(restApiId=self._api_physical_id, description="Created by SAM Sync")
        )
        new_dep_id = response_dep.get("id")
        LOG.debug("%sCreate Deployment Result: %s", self.log_prefix, response_dep)
        return new_dep_id

    def _collect_stages(self) -> Set[str]:
        """
        Collect all stages needed to be updated

        Returns
        ----------
        Set[str]: The set of stage names to be updated
        """
        # Get the stage name associated with the previous deployment and update stage
        # Stage needs to be flushed so that new changes will be visible immediately
        api_resource = get_resource_by_id(self._stacks, ResourceIdentifier(self._api_identifier))
        stage_resources = get_resource_ids_by_type(self._stacks, AWS_APIGATEWAY_STAGE)
        deployment_resources = get_resource_ids_by_type(self._stacks, AWS_APIGATEWAY_DEPLOYMENT)

        stages = set()
        # If it is a SAM resource, get the StageName property
        if api_resource:
            if api_resource.get("Type") == AWS_SERVERLESS_API:
                # The customer defined stage name
                stage_name = api_resource.get("Properties", {}).get("StageName")
                if stage_name:
                    stages.add(cast(str, stage_name))

                # The stage called "Stage"
                if stage_name != "Stage":
                    response_sta = cast(Dict, self._api_client.get_stages(restApiId=self._api_physical_id))
                    for item in response_sta.get("item"):  # type: ignore
                        if item.get("stageName") == "Stage":
                            stages.add("Stage")

        # For both SAM and ApiGateway resource, check if any refs from stage resources
        for stage_resource in stage_resources:
            # RestApiId is a required field in stage
            stage_dict = get_resource_by_id(self._stacks, stage_resource)
            if not stage_dict:
                continue
            rest_api_id = stage_dict.get("Properties", {}).get("RestApiId")
            dep_id = stage_dict.get("Properties", {}).get("DeploymentId")
            # If the stage doesn't have a deployment associated then no need to update
            if dep_id is None:
                continue
            # If the stage's deployment ID is not static and the rest API ID matchs, then update
            for deployment_resource in deployment_resources:
                if deployment_resource.resource_iac_id == dep_id and rest_api_id == self._api_identifier:
                    stages.add(cast(str, stage_dict.get("Properties", {}).get("StageName")))
                    break

        return stages

    def _update_stages(self, stages: Set[str], deployment_id: Optional[str]) -> Set[str]:
        """
        Update all the relevant stages

        Parameters
        ----------
        stages: Set[str]
            The set of stage names to be updated
        deployment_id: Optional[str]
            The newly created deployment ID to be used in the stages
        Returns
        ----------
        Set[str]: A set of previous deployment IDs to be cleaned up
        """
        prev_dep_ids = set()
        for stage in stages:
            # Collects previous deployment IDs to clean up
            response_get = cast(Dict, self._api_client.get_stage(restApiId=self._api_physical_id, stageName=stage))
            prev_dep_id = response_get.get("deploymentId")
            if prev_dep_id:
                prev_dep_ids.add(cast(str, prev_dep_id))

            # Updates the stage with newest deployment
            LOG.debug("%sTrying to update the stage %s through client", self.log_prefix, stage)
            response_upd = cast(
                Dict,
                self._api_client.update_stage(
                    restApiId=self._api_physical_id,
                    stageName=stage,
                    patchOperations=[{"op": "replace", "path": "/deploymentId", "value": deployment_id}],
                ),
            )
            LOG.debug("%sUpdate Stage Result: %s", self.log_prefix, response_upd)

            # Flushes the cache so that the changes are calleable
            self._api_client.flush_stage_cache(restApiId=self._api_physical_id, stageName=stage)
            self._api_client.flush_stage_authorizers_cache(restApiId=self._api_physical_id, stageName=stage)
        return prev_dep_ids

    def _delete_deployments(self, prev_deployment_ids: Set[str]) -> None:
        """
        Delete the previous deployment

        Parameters
        ----------
        prev_deployment_ids: Set[str]
            A set of previous deployment IDs to be cleaned up
        """
        for prev_dep_id in prev_deployment_ids:
            try:
                LOG.debug("%sTrying to delete the previous deployment %s through client", self.log_prefix, prev_dep_id)
                response_del = cast(
                    Dict, self._api_client.delete_deployment(restApiId=self._api_physical_id, deploymentId=prev_dep_id)
                )
                LOG.debug("%sDelete Deployment Result: %s", self.log_prefix, response_del)
            except ClientError:
                LOG.warning(
                    Colored().yellow(
                        "Delete deployment for %s failed, it may be due to the it being used by another stage. \
please check the console to see if you have other stages that needs to be updated."
                    ),
                    prev_dep_id,
                )



"""
This file contains puller implementations for XRay
"""
import logging
import time
from datetime import datetime
from itertools import zip_longest
from typing import Any, Dict, List, Optional, Union

from botocore.exceptions import ClientError

from samcli.lib.observability.observability_info_puller import ObservabilityEventConsumer, ObservabilityPuller
from samcli.lib.observability.xray_traces.xray_events import XRayTraceEvent
from samcli.lib.utils.time import to_datetime, to_timestamp

LOG = logging.getLogger(__name__)


class AbstractXRayPuller(ObservabilityPuller):
    def __init__(
        self,
        max_retries: int = 1000,
        poll_interval: int = 1,
    ):
        """
        Parameters
        ----------
        max_retries : int
            Optional maximum number of retries which can be used to pull information. Default value is 1000
        poll_interval : int
            Optional interval value that will be used to wait between calls in tail operation. Default value is 1
        """
        self._max_retries = max_retries
        self._poll_interval = poll_interval
        self._had_data = False
        self.latest_event_time = 0

    def tail(self, start_time: Optional[datetime] = None, filter_pattern: Optional[str] = None):
        if start_time:
            self.latest_event_time = to_timestamp(start_time)

        counter = self._max_retries
        while counter > 0 and not self.cancelled:
            LOG.debug("Tailing XRay traces starting at %s", self.latest_event_time)

            counter -= 1
            try:
                self.load_time_period(to_datetime(self.latest_event_time), datetime.utcnow())
            except ClientError as err:
                error_code = err.response.get("Error", {}).get("Code")
                if error_code == "ThrottlingException":
                    # if throttled, increase poll interval by 1 second each time
                    if self._poll_interval == 1:
                        self._poll_interval += 1
                    else:
                        self._poll_interval **= 2
                    LOG.warning(
                        "Throttled by XRay API, increasing the poll interval time to %s seconds",
                        self._poll_interval,
                    )
                else:
                    # if exception is other than throttling re-raise
                    LOG.error("Failed while fetching new AWS X-Ray events", exc_info=err)
                    raise err

            if self._had_data:
                counter = self._max_retries
                self.latest_event_time += 1
                self._had_data = False

            time.sleep(self._poll_interval)


class XRayTracePuller(AbstractXRayPuller):
    """
    ObservabilityPuller implementation which pulls XRay trace information by summarizing XRay traces first
    and then getting them as a batch later.
    """

    def __init__(
        self, xray_client: Any, consumer: ObservabilityEventConsumer, max_retries: int = 1000, poll_interval: int = 1
    ):
        """
        Parameters
        ----------
        xray_client : boto3.client
            XRay boto3 client instance
        consumer :  ObservabilityEventConsumer
            Consumer instance which will process pulled events
        max_retries : int
            Optional maximum number of retries which can be used to pull information. Default value is 1000
        poll_interval : int
            Optional interval value that will be used to wait between calls in tail operation. Default value is 1
        """
        super().__init__(max_retries, poll_interval)
        self.xray_client = xray_client
        self.consumer = consumer
        # Previous trace ID is a dictionary that contains the following information: {trace_id: trace_revision,}
        self._previous_trace_ids: Dict = {}

    def load_time_period(
        self,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        filter_pattern: Optional[str] = None,
    ):
        kwargs = {"TimeRangeType": "TraceId", "StartTime": start_time, "EndTime": end_time}

        # first, collect all trace ids in given period
        trace_ids = {}
        LOG.debug("Fetching XRay trace summaries %s", kwargs)
        result_paginator = self.xray_client.get_paginator("get_trace_summaries")
        result_iterator = result_paginator.paginate(**kwargs)
        for result in result_iterator:
            trace_summaries = result.get("TraceSummaries", [])
            for trace_summary in trace_summaries:
                trace_id = trace_summary.get("Id", None)
                trace_revision = int(trace_summary.get("Revision", 0))
                is_partial = trace_summary.get("IsPartial", False)
                if not is_partial:
                    if trace_id not in self._previous_trace_ids or trace_revision > self._previous_trace_ids.get(
                        trace_id, 0
                    ):
                        trace_ids[trace_id] = trace_revision
                        self._previous_trace_ids[trace_id] = trace_revision

        # now load collected events
        self.load_events(trace_ids)

    def load_events(self, event_ids: Union[List[Any], Dict]):
        # event_ids have trace ID as key and revision number as value
        if not event_ids:
            LOG.debug("Nothing to fetch, empty event_id dict given (%s)", event_ids)
            return

        if isinstance(event_ids, dict):
            # xray client only accepts 5 items at max, so create batches of 5 element arrays
            event_batches = zip_longest(*([iter(event_ids.keys())] * 5))
        else:
            event_batches = zip_longest(*([iter(event_ids)] * 5))

        for event_batch in event_batches:
            kwargs: Dict[str, Any] = {"TraceIds": list(filter(None, event_batch))}
            result_paginator = self.xray_client.get_paginator("batch_get_traces")
            result_iterator = result_paginator.paginate(**kwargs)
            for result in result_iterator:
                traces = result.get("Traces", [])

                if not traces:
                    LOG.debug("No event found with given trace ids %s", str(event_ids))

                for trace in traces:
                    self._had_data = True
                    trace_id = trace.get("Id", None)
                    if isinstance(event_ids, dict):
                        xray_trace_event = XRayTraceEvent(trace, event_ids.get(trace_id, None))
                    else:
                        xray_trace_event = XRayTraceEvent(trace)

                    # update latest fetched event
                    latest_event_time = xray_trace_event.get_latest_event_time()
                    if latest_event_time > self.latest_event_time:
                        self.latest_event_time = latest_event_time

                    self.consumer.consume(xray_trace_event)



"""
Client for uploading packaged artifacts to s3
"""

# Copyright 2012-2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import logging
import os
import sys
import threading
from collections import abc
from typing import Any, Dict, Optional, cast
from urllib.parse import parse_qs, urlparse

import botocore
import botocore.exceptions
from boto3.s3 import transfer

from samcli.commands.package.exceptions import BucketNotSpecifiedError, NoSuchBucketError
from samcli.lib.package.local_files_utils import get_uploaded_s3_object_name

LOG = logging.getLogger(__name__)


class S3Uploader:
    """
    Class to upload objects to S3 bucket that use versioning. If bucket
    does not already use versioning, this class will turn on versioning.
    """

    @property
    def artifact_metadata(self):
        """
        Metadata to attach to the object(s) uploaded by the uploader.
        """
        return self._artifact_metadata

    @artifact_metadata.setter
    def artifact_metadata(self, val):
        if val is not None and not isinstance(val, abc.Mapping):
            raise TypeError("Artifact metadata should be in dict type")
        self._artifact_metadata = val

    def __init__(
        self,
        s3_client: Any,
        bucket_name: str,
        prefix: Optional[str] = None,
        kms_key_id: Optional[str] = None,
        force_upload: bool = False,
        no_progressbar: bool = False,
    ):
        self.s3 = s3_client
        self.bucket_name = bucket_name
        self.prefix = prefix
        self.kms_key_id = kms_key_id or None
        self.force_upload = force_upload
        self.no_progressbar = no_progressbar
        self.transfer_manager = transfer.create_transfer_manager(self.s3, transfer.TransferConfig())

        self._artifact_metadata = None

    def upload(self, file_name: str, remote_path: str) -> str:
        """
        Uploads given file to S3
        :param file_name: Path to the file that will be uploaded
        :param remote_path:  be uploaded
        :return: VersionId of the latest upload
        """

        if self.prefix:
            remote_path = "{0}/{1}".format(self.prefix, remote_path)

        # Check if a file with same data exists
        if not self.force_upload and self.file_exists(remote_path):
            LOG.info("File with same data already exists at %s, skipping upload", remote_path)
            return self.make_url(remote_path)

        try:

            # Default to regular server-side encryption unless customer has
            # specified their own KMS keys
            additional_args = {"ServerSideEncryption": "AES256"}

            if self.kms_key_id:
                additional_args["ServerSideEncryption"] = "aws:kms"
                additional_args["SSEKMSKeyId"] = self.kms_key_id

            if self.artifact_metadata:
                additional_args["Metadata"] = self.artifact_metadata

            if not self.bucket_name:
                raise BucketNotSpecifiedError()

            if not self.no_progressbar:
                print_progress_callback = ProgressPercentage(file_name, remote_path)
                future = self.transfer_manager.upload(
                    file_name, self.bucket_name, remote_path, additional_args, [print_progress_callback]
                )
            else:
                future = self.transfer_manager.upload(file_name, self.bucket_name, remote_path, additional_args)
            future.result()

            return self.make_url(remote_path)

        except botocore.exceptions.ClientError as ex:
            error_code = ex.response["Error"]["Code"]
            if error_code == "NoSuchBucket":
                raise NoSuchBucketError(bucket_name=self.bucket_name) from ex
            raise ex

    def upload_with_dedup(
        self, file_name: str, extension: Optional[str] = None, precomputed_md5: Optional[str] = None
    ) -> str:
        """
        Makes and returns name of the S3 object based on the file's MD5 sum

        :param file_name: file to upload
        :param extension: String of file extension to append to the object
        :param precomputed_md5: Specified md5 hash for the file to be uploaded.
        :return: S3 URL of the uploaded object
        """

        # This construction of remote_path is critical to preventing duplicate
        # uploads of same object. Uploader will check if the file exists in S3
        # and re-upload only if necessary. So the template points to same file
        # in multiple places, this will upload only once
        remote_path = get_uploaded_s3_object_name(
            precomputed_md5=precomputed_md5, file_path=file_name, extension=extension
        )
        return self.upload(file_name, remote_path)

    def delete_artifact(self, remote_path: str, is_key: bool = False) -> bool:
        """
        Deletes a given file from S3
        :param remote_path: Path to the file that will be deleted
        :param is_key: If the given remote_path is the key or a file_name

        :return: metadata dict of the deleted object
        """
        try:
            if not self.bucket_name:
                LOG.error("Bucket not specified")
                raise BucketNotSpecifiedError()

            key = remote_path
            if self.prefix and not is_key:
                key = "{0}/{1}".format(self.prefix, remote_path)

            # Deleting Specific file with key
            if self.file_exists(remote_path=key):
                LOG.info("\t- Deleting S3 object with key %s", key)
                self.s3.delete_object(Bucket=self.bucket_name, Key=key)
                LOG.debug("Deleted s3 object with key %s successfully", key)
                return True

            # Given s3 object key does not exist
            LOG.debug("Could not find the S3 file with the key %s", key)
            LOG.info("\t- Could not find and delete the S3 object with the key %s", key)
            return False

        except botocore.exceptions.ClientError as ex:
            error_code = ex.response["Error"]["Code"]
            if error_code == "NoSuchBucket":
                LOG.error("Provided bucket %s does not exist ", self.bucket_name)
                raise NoSuchBucketError(bucket_name=self.bucket_name) from ex
            raise ex

    def delete_prefix_artifacts(self):
        """
        Deletes all the files from the prefix in S3
        """
        if not self.bucket_name:
            LOG.error("Bucket not specified")
            raise BucketNotSpecifiedError()
        if self.prefix:
            # Note: list_objects_v2 api uses prefix to fetch the keys that begin with the prefix
            # To restrict fetching files with exact prefix self.prefix, "/" is used below.
            response = self.s3.list_objects_v2(Bucket=self.bucket_name, Prefix=self.prefix + "/")
            prefix_files = response.get("Contents", [])
            for obj in prefix_files:
                self.delete_artifact(obj["Key"], True)

    def file_exists(self, remote_path: str) -> bool:
        """
        Check if the file we are trying to upload already exists in S3

        :param remote_path:
        :return: True, if file exists. False, otherwise
        """

        try:
            # Find the object that matches this ETag
            if not self.bucket_name:
                raise BucketNotSpecifiedError()
            self.s3.head_object(Bucket=self.bucket_name, Key=remote_path)
            return True
        except botocore.exceptions.ClientError:
            # Either File does not exist or we are unable to get
            # this information.
            return False

    def make_url(self, obj_path: str) -> str:
        if not self.bucket_name:
            raise BucketNotSpecifiedError()
        return "s3://{0}/{1}".format(self.bucket_name, obj_path)

    def to_path_style_s3_url(self, key: str, version: Optional[str] = None) -> str:
        """
        This link describes the format of Path Style URLs
        http://docs.aws.amazon.com/AmazonS3/latest/dev/UsingBucket.html#access-bucket-intro
        """
        base = self.s3.meta.endpoint_url
        result = "{0}/{1}/{2}".format(base, self.bucket_name, key)
        if version:
            result = "{0}?versionId={1}".format(result, version)

        return result

    def get_version_of_artifact(self, s3_url: str) -> str:
        """
        Returns version information of the S3 object that is given as S3 URL
        """
        parsed_s3_url = self.parse_s3_url(s3_url)
        s3_bucket = parsed_s3_url["Bucket"]
        s3_key = parsed_s3_url["Key"]
        s3_object_tagging = self.s3.get_object_tagging(Bucket=s3_bucket, Key=s3_key)
        LOG.debug("S3 Object (%s) tagging information %s", s3_url, s3_object_tagging)
        s3_object_version_id = s3_object_tagging["VersionId"]
        return cast(str, s3_object_version_id)

    @staticmethod
    def parse_s3_url(
        url: Any,
        bucket_name_property: str = "Bucket",
        object_key_property: str = "Key",
        version_property: Optional[str] = None,
    ) -> Dict:
        if isinstance(url, str) and url.startswith("s3://"):

            return S3Uploader._parse_s3_format_url(
                url=url,
                bucket_name_property=bucket_name_property,
                object_key_property=object_key_property,
                version_property=version_property,
            )

        if isinstance(url, str) and url.startswith("https://s3"):
            return S3Uploader._parse_path_style_s3_url(
                url=url, bucket_name_property=bucket_name_property, object_key_property=object_key_property
            )

        raise ValueError("URL given to the parse method is not a valid S3 url {0}".format(url))

    @staticmethod
    def _parse_s3_format_url(
        url: Any,
        bucket_name_property: str = "Bucket",
        object_key_property: str = "Key",
        version_property: Optional[str] = None,
    ) -> Dict:
        """
        Method for parsing s3 urls that begin with s3://
        e.g. s3://bucket/key
        """
        parsed = urlparse(url)
        query = parse_qs(parsed.query)
        if parsed.netloc and parsed.path:
            result = dict()
            result[bucket_name_property] = parsed.netloc
            result[object_key_property] = parsed.path.lstrip("/")

            # If there is a query string that has a single versionId field,
            # set the object version and return
            if version_property is not None and "versionId" in query and len(query["versionId"]) == 1:
                result[version_property] = query["versionId"][0]

            return result

        raise ValueError("URL given to the parse method is not a valid S3 url {0}".format(url))

    @staticmethod
    def _parse_path_style_s3_url(
        url: Any,
        bucket_name_property: str = "Bucket",
        object_key_property: str = "Key",
    ) -> Dict:
        """
        Static method for parsing path style s3 urls.
        e.g. https://s3.us-east-1.amazonaws.com/bucket/key
        """
        parsed = urlparse(url)
        result = dict()
        # parsed.path would point to /bucket/key
        if parsed.path:
            s3_bucket_key = parsed.path.split("/", 2)[1:]

            result[bucket_name_property] = s3_bucket_key[0]
            result[object_key_property] = s3_bucket_key[1]

            return result
        raise ValueError("URL given to the parse method is not a valid S3 url {0}".format(url))


class ProgressPercentage:
    # This class was copied directly from S3Transfer docs

    def __init__(self, filename, remote_path):
        self._filename = filename
        self._remote_path = remote_path
        self._size = os.path.getsize(filename)
        self._seen_so_far = 0
        self._lock = threading.Lock()

    def on_progress(self, bytes_transferred, **kwargs):

        # To simplify we'll assume this is hooked up
        # to a single filename.
        with self._lock:
            self._seen_so_far += bytes_transferred
            percentage = (self._seen_so_far / self._size) * 100  # noqa: PLR2004
            sys.stderr.write(
                "\r\tUploading to %s  %s / %s  (%.2f%%)"
                % (self._remote_path, self._seen_so_far, self._size, percentage)
            )
            sys.stderr.flush()
            if int(percentage) == 100:  # noqa: PLR2004
                sys.stderr.write("\n")



"""
CLI configuration decorator to use TOML configuration files for click commands.
"""

# This section contains code copied and modified from
# [click_config_file](https://github.com/phha/click_config_file/blob/master/click_config_file.py)
# SPDX-License-Identifier: MIT

import functools
import logging
import os
from pathlib import Path

import click

from samcli.cli.context import get_cmd_names
from samcli.commands.exceptions import ConfigException
from samcli.lib.config.samconfig import DEFAULT_CONFIG_FILE_NAME, DEFAULT_ENV, SamConfig

__all__ = ("TomlProvider", "configuration_option", "get_ctx_defaults")

LOG = logging.getLogger(__name__)


class TomlProvider:
    """
    A parser for toml configuration files
    """

    def __init__(self, section=None, cmd_names=None):
        """
        The constructor for TomlProvider class
        :param section: section defined in the configuration file nested within `cmd`
        :param cmd_names: cmd_name defined in the configuration file
        """
        self.section = section
        self.cmd_names = cmd_names

    def __call__(self, config_path, config_env, cmd_names):
        """
        Get resolved config based on the `file_path` for the configuration file,
        `config_env` targeted inside the config file and corresponding `cmd_name`
        as denoted by `click`.

        :param config_path: The path of configuration file.
        :param config_env: The name of the sectional config_env within configuration file.
        :param list cmd_names: sam command name as defined by click
        :returns dictionary containing the configuration parameters under specified config_env
        """

        resolved_config = {}

        # Use default sam config file name if config_path only contain the directory
        config_file_path = (
            Path(os.path.abspath(config_path)) if config_path else Path(os.getcwd(), DEFAULT_CONFIG_FILE_NAME)
        )
        config_file_name = config_file_path.name
        config_file_dir = config_file_path.parents[0]

        samconfig = SamConfig(config_file_dir, config_file_name)

        # Enable debug level logging by environment variable "SAM_DEBUG"
        if os.environ.get("SAM_DEBUG", "").lower() == "true":
            LOG.setLevel(logging.DEBUG)

        LOG.debug("Config file location: %s", samconfig.path())

        if not samconfig.exists():
            LOG.debug("Config file '%s' does not exist", samconfig.path())
            return resolved_config

        if not self.cmd_names:
            self.cmd_names = cmd_names

        try:
            LOG.debug(
                "Loading configuration values from [%s.%s.%s] (env.command_name.section) in config file at '%s'...",
                config_env,
                self.cmd_names,
                self.section,
                samconfig.path(),
            )

            # NOTE(TheSriram): change from tomlkit table type to normal dictionary,
            # so that click defaults work out of the box.
            resolved_config = dict(samconfig.get_all(self.cmd_names, self.section, env=config_env).items())
            LOG.debug("Configuration values successfully loaded.")
            LOG.debug("Configuration values are: %s", resolved_config)

        except KeyError as ex:
            LOG.debug(
                "Error reading configuration from [%s.%s.%s] (env.command_name.section) "
                "in configuration file at '%s' with : %s",
                config_env,
                self.cmd_names,
                self.section,
                samconfig.path(),
                str(ex),
            )

        except Exception as ex:
            LOG.debug("Error reading configuration file: %s %s", samconfig.path(), str(ex))
            raise ConfigException(f"Error reading configuration: {ex}") from ex

        return resolved_config


def configuration_callback(cmd_name, option_name, saved_callback, provider, ctx, param, value):
    """
    Callback for reading the config file.

    Also takes care of calling user specified custom callback afterwards.

    :param cmd_name: `sam` command name derived from click.
    :param option_name: The name of the option. This is used for error messages.
    :param saved_callback: User-specified callback to be called later.
    :param provider:  A callable that parses the configuration file and returns a dictionary
        of the configuration parameters. Will be called as
        `provider(file_path, config_env, cmd_name)`.
    :param ctx: Click context
    :param param: Click parameter
    :param value: Specified value for config_env
    :returns specified callback or the specified value for config_env.
    """

    # ctx, param and value are default arguments for click specified callbacks.
    ctx.default_map = ctx.default_map or {}
    cmd_name = cmd_name or ctx.info_name
    param.default = None
    config_env_name = ctx.params.get("config_env") or DEFAULT_ENV

    config_file = ctx.params.get("config_file") or DEFAULT_CONFIG_FILE_NAME
    config_dir = getattr(ctx, "samconfig_dir", None) or os.getcwd()
    # If --config-file is an absolute path, use it, if not, start from config_dir
    config_file_path = config_file if os.path.isabs(config_file) else os.path.join(config_dir, config_file)
    if config_file and config_file != DEFAULT_CONFIG_FILE_NAME and not Path(config_file_path).absolute().is_file():
        error_msg = f"Config file {config_file} does not exist or could not be read!"
        LOG.debug(error_msg)
        raise ConfigException(error_msg)

    config = get_ctx_defaults(
        cmd_name,
        provider,
        ctx,
        config_env_name=config_env_name,
        config_file=config_file_path,
    )
    ctx.default_map.update(config)

    return saved_callback(ctx, param, config_env_name) if saved_callback else config_env_name


def get_ctx_defaults(cmd_name, provider, ctx, config_env_name, config_file=None):
    """
    Get the set of the parameters that are needed to be set into the click command.
    This function also figures out the command name by looking up current click context's parent
    and constructing the parsed command name that is used in default configuration file.
    If a given cmd_name is start-api, the parsed name is "local_start_api".
    provider is called with `config_file`, `config_env_name` and `parsed_cmd_name`.

    :param cmd_name: `sam` command name
    :param provider: provider to be called for reading configuration file
    :param ctx: Click context
    :param config_env_name: config-env within configuration file, sam configuration file will be relative to the
                            supplied original template if its path is not specified
    :param config_file: configuration file name
    :return: dictionary of defaults for parameters
    """

    return provider(config_file, config_env_name, get_cmd_names(cmd_name, ctx))


def configuration_option(*param_decls, **attrs):
    # pylint does not understand the docstring with the presence of **attrs
    # pylint: disable=missing-param-doc,differing-param-doc
    """
    Adds configuration file support to a click application.

    NOTE: This decorator should be added to the top of parameter chain, right below click.command, before
          any options are declared.

    Example:
        >>> @click.command("hello")
            @configuration_option(provider=TomlProvider(section="parameters"))
            @click.option('--name', type=click.String)
            def hello(name):
                print("Hello " + name)

    This will create a hidden click option whose callback function loads configuration parameters from default
    configuration environment [default] in default configuration file [samconfig.toml] in the template file
    directory.
    :param preconfig_decorator_list: A list of click option decorator which need to place before this function. For
        exmple, if we want to add option "--config-file" and "--config-env" to allow customized configuration file
        and configuration environment, we will use configuration_option as below:
        @configuration_option(
            preconfig_decorator_list=[decorator_customize_config_file, decorator_customize_config_env],
            provider=TomlProvider(section=CONFIG_SECTION),
        )
        By default, we enable these two options.
    :param provider: A callable that parses the configuration file and returns a dictionary
        of the configuration parameters. Will be called as
        `provider(file_path, config_env, cmd_name)
    """

    def decorator_configuration_setup(f):
        configuration_setup_params = ()
        configuration_setup_attrs = {}
        configuration_setup_attrs[
            "help"
        ] = "This is a hidden click option whose callback function loads configuration parameters."
        configuration_setup_attrs["is_eager"] = True
        configuration_setup_attrs["expose_value"] = False
        configuration_setup_attrs["hidden"] = True
        configuration_setup_attrs["type"] = click.STRING
        provider = attrs.pop("provider")
        saved_callback = attrs.pop("callback", None)
        partial_callback = functools.partial(configuration_callback, None, None, saved_callback, provider)
        configuration_setup_attrs["callback"] = partial_callback
        return click.option(*configuration_setup_params, **configuration_setup_attrs)(f)

    def composed_decorator(decorators):
        def decorator(f):
            for deco in decorators:
                f = deco(f)
            return f

        return decorator

    # Compose decorators here to make sure the context parameters are updated before callback function
    decorator_list = [decorator_configuration_setup]
    pre_config_decorators = attrs.pop(
        "preconfig_decorator_list", [decorator_customize_config_file, decorator_customize_config_env]
    )
    for decorator in pre_config_decorators:
        decorator_list.append(decorator)
    return composed_decorator(decorator_list)


def decorator_customize_config_file(f):
    """
    CLI option to customize configuration file name. By default it is 'samconfig.toml' in project directory.
    Ex: --config-file samconfig.toml
    :param f: Callback function passed by Click
    :return: Callback function
    """
    config_file_attrs = {}
    config_file_param_decls = ("--config-file",)
    config_file_attrs["help"] = (
        "The path and file name of the configuration file containing default parameter values to use. "
        "Its default value is 'samconfig.toml' in project directory. For more information about configuration files, "
        "see: "
        "https://docs.aws.amazon.com/serverless-application-model/latest/developerguide/serverless-sam-cli-config.html."
    )
    config_file_attrs["default"] = "samconfig.toml"
    config_file_attrs["is_eager"] = True
    config_file_attrs["required"] = False
    config_file_attrs["type"] = click.STRING
    return click.option(*config_file_param_decls, **config_file_attrs)(f)


def decorator_customize_config_env(f):
    """
    CLI option to customize configuration environment name. By default it is 'default'.
    Ex: --config-env default
    :param f: Callback function passed by Click
    :return: Callback function
    """
    config_env_attrs = {}
    config_env_param_decls = ("--config-env",)
    config_env_attrs["help"] = (
        "The environment name specifying the default parameter values in the configuration file to use. "
        "Its default value is 'default'. For more information about configuration files, see: "
        "https://docs.aws.amazon.com/serverless-application-model/latest/developerguide/serverless-sam-cli-config.html."
    )
    config_env_attrs["default"] = "default"
    config_env_attrs["is_eager"] = True
    config_env_attrs["required"] = False
    config_env_attrs["type"] = click.STRING
    return click.option(*config_env_param_decls, **config_env_attrs)(f)


# End section copied from
# [click_config_file](https://github.com/phha/click_config_file/blob/master/click_config_file.py)



"""
Utilities for local files handling.
"""
import os
import tempfile
import uuid
from contextlib import contextmanager
from typing import Optional

from samcli.lib.utils.hash import file_checksum


@contextmanager
def mktempfile():
    directory = tempfile.gettempdir()
    filename = os.path.join(directory, uuid.uuid4().hex)

    try:
        with open(filename, "w+") as handle:
            yield handle
    finally:
        if os.path.exists(filename):
            os.remove(filename)


def get_uploaded_s3_object_name(
    precomputed_md5: Optional[str] = None,
    file_content: Optional[str] = None,
    file_path: Optional[str] = None,
    extension: Optional[str] = None,
) -> str:
    """
    Generate the file name that will be used while creating the S3 Object based on the file hash value.
    This method expect either the precomuted hash value of the file, or the file content, or the file path

    Parameters
    ----------
    precomputed_md5: str
        the precomputed hash value of the file.
    file_content : str
        The file content to be uploaded to S3.
    file_path : str
        The file path to be uploaded to S3
    extension : str
        The file extension in S3
    Returns
    -------
    str
        The generated S3 Object name
    """
    if precomputed_md5:
        filemd5 = precomputed_md5
    elif file_content:
        with mktempfile() as temp_file:
            temp_file.write(file_content)
            temp_file.flush()
            filemd5 = file_checksum(temp_file.name)
    elif file_path:
        filemd5 = file_checksum(file_path)
    else:
        raise Exception("Either File Content, File Path, or Precomputed Hash should has a value")

    if extension:
        filemd5 = filemd5 + "." + extension

    return filemd5



"""
Contains sync flow implementation for Auto Dependency Layer
"""
import hashlib
import logging
import os
import tempfile
import uuid
from typing import TYPE_CHECKING, Dict, List, Optional, cast

from samcli.lib.bootstrap.nested_stack.nested_stack_builder import NestedStackBuilder
from samcli.lib.bootstrap.nested_stack.nested_stack_manager import NestedStackManager
from samcli.lib.build.build_graph import BuildGraph
from samcli.lib.package.utils import make_zip_with_lambda_permissions
from samcli.lib.providers.provider import Function, Stack
from samcli.lib.providers.sam_function_provider import SamFunctionProvider
from samcli.lib.sync.exceptions import (
    InvalidRuntimeDefinitionForFunction,
    MissingFunctionBuildDefinition,
    NoLayerVersionsFoundError,
)
from samcli.lib.sync.flows.layer_sync_flow import AbstractLayerSyncFlow
from samcli.lib.sync.flows.zip_function_sync_flow import ZipFunctionSyncFlow
from samcli.lib.sync.sync_flow import SyncFlow
from samcli.lib.utils.hash import file_checksum

if TYPE_CHECKING:  # pragma: no cover
    from samcli.commands.build.build_context import BuildContext
    from samcli.commands.deploy.deploy_context import DeployContext
    from samcli.commands.sync.sync_context import SyncContext

LOG = logging.getLogger(__name__)


class AutoDependencyLayerSyncFlow(AbstractLayerSyncFlow):
    """
    Auto Dependency Layer, Layer Sync flow.
    It creates auto dependency layer files out of function dependencies, and syncs layer code and then updates
    the function configuration with new layer version

    This flow is not instantiated from factory method, please see AutoDependencyLayerParentSyncFlow
    """

    _function_identifier: str
    _build_graph: Optional[BuildGraph]

    def __init__(
        self,
        function_identifier: str,
        build_graph: BuildGraph,
        build_context: "BuildContext",
        deploy_context: "DeployContext",
        sync_context: "SyncContext",
        physical_id_mapping: Dict[str, str],
        stacks: List[Stack],
    ):
        super().__init__(
            NestedStackBuilder.get_layer_logical_id(function_identifier),
            build_context,
            deploy_context,
            sync_context,
            physical_id_mapping,
            stacks,
        )
        self._function_identifier = function_identifier
        self._build_graph = build_graph

    def set_up(self) -> None:
        super().set_up()

        # find layer's physical id
        layer_name = NestedStackBuilder.get_layer_name(self._deploy_context.stack_name, self._function_identifier)
        layer_versions = self._lambda_client.list_layer_versions(LayerName=layer_name).get("LayerVersions", [])
        if not layer_versions:
            raise NoLayerVersionsFoundError(layer_name)
        self._layer_arn = layer_versions[0].get("LayerVersionArn").rsplit(":", 1)[0]

    def gather_resources(self) -> None:
        function_build_definitions = cast(BuildGraph, self._build_graph).get_function_build_definitions()
        if not function_build_definitions:
            raise MissingFunctionBuildDefinition(self._function_identifier)

        self._artifact_folder = NestedStackManager.update_layer_folder(
            self._build_context.build_dir,
            function_build_definitions[0].dependencies_dir,
            self._layer_identifier,
            self._function_identifier,
            self._get_compatible_runtimes()[0],
        )
        zip_file_path = os.path.join(tempfile.gettempdir(), "data-" + uuid.uuid4().hex)
        self._zip_file = make_zip_with_lambda_permissions(zip_file_path, self._artifact_folder)
        self._local_sha = file_checksum(cast(str, self._zip_file), hashlib.sha256())

    def _get_dependent_functions(self) -> List[Function]:
        function = SamFunctionProvider(cast(List[Stack], self._stacks)).get(self._function_identifier)
        return [function] if function else []

    def _get_compatible_runtimes(self) -> List[str]:
        function = SamFunctionProvider(cast(List[Stack], self._stacks)).get(self._function_identifier)
        if not function or not function.runtime:
            raise InvalidRuntimeDefinitionForFunction(self._function_identifier)
        return [function.runtime]


class AutoDependencyLayerParentSyncFlow(ZipFunctionSyncFlow):
    """
    Parent sync flow for auto dependency layer

    It builds function with regular ZipFunctionSyncFlow, and then adds _AutoDependencyLayerSyncFlow to start syncing
    dependency layer.
    """

    def gather_dependencies(self) -> List[SyncFlow]:
        """
        Return auto dependency layer sync flow along with parent dependencies
        """
        parent_dependencies = super().gather_dependencies()

        function_build_definitions = cast(BuildGraph, self._build_graph).get_function_build_definitions()
        if not function_build_definitions:
            raise MissingFunctionBuildDefinition(self._function.name)

        # don't queue up auto dependency layer, if dependencies are not changes
        need_dependency_layer_sync = function_build_definitions[0].download_dependencies
        if need_dependency_layer_sync:
            parent_dependencies.append(
                AutoDependencyLayerSyncFlow(
                    self._function_identifier,
                    cast(BuildGraph, self._build_graph),
                    self._build_context,
                    self._deploy_context,
                    self._sync_context,
                    self._physical_id_mapping,
                    cast(List[Stack], self._stacks),
                )
            )
        return parent_dependencies

    @staticmethod
    def _combine_dependencies() -> bool:
        return False



"""CLI command for "sync" command."""
import logging
import os
from typing import TYPE_CHECKING, List, Optional, Set, Tuple

import click

from samcli.cli.cli_config_file import TomlProvider, configuration_option
from samcli.cli.context import Context
from samcli.cli.main import aws_creds_options, pass_context, print_cmdline_args
from samcli.cli.main import common_options as cli_framework_options
from samcli.commands._utils.cdk_support_decorators import unsupported_command_cdk
from samcli.commands._utils.click_mutex import ClickMutex
from samcli.commands._utils.command_exception_handler import command_exception_handler
from samcli.commands._utils.constants import (
    DEFAULT_BUILD_DIR,
    DEFAULT_BUILD_DIR_WITH_AUTO_DEPENDENCY_LAYER,
    DEFAULT_CACHE_DIR,
)
from samcli.commands._utils.options import (
    base_dir_option,
    capabilities_option,
    image_repositories_option,
    image_repository_option,
    kms_key_id_option,
    metadata_option,
    notification_arns_option,
    parameter_override_option,
    role_arn_option,
    s3_bucket_option,
    s3_prefix_option,
    stack_name_option,
    tags_option,
    template_option_without_build,
    use_container_build_option,
)
from samcli.commands.build.command import _get_mode_value_from_envvar
from samcli.commands.sync.sync_context import SyncContext
from samcli.lib.bootstrap.bootstrap import manage_stack
from samcli.lib.build.bundler import EsbuildBundlerManager
from samcli.lib.cli_validation.image_repository_validation import image_repository_validation
from samcli.lib.providers.provider import (
    ResourceIdentifier,
    get_all_resource_ids,
    get_unique_resource_ids,
)
from samcli.lib.providers.sam_stack_provider import SamLocalStackProvider
from samcli.lib.sync.sync_flow_executor import SyncFlowExecutor
from samcli.lib.sync.sync_flow_factory import SyncCodeResources, SyncFlowFactory
from samcli.lib.sync.watch_manager import WatchManager
from samcli.lib.telemetry.event import EventTracker, track_long_event
from samcli.lib.telemetry.metric import track_command, track_template_warnings
from samcli.lib.utils.colors import Colored
from samcli.lib.utils.version_checker import check_newer_version
from samcli.lib.warnings.sam_cli_warning import CodeDeployConditionWarning, CodeDeployWarning

if TYPE_CHECKING:  # pragma: no cover
    from samcli.commands.build.build_context import BuildContext
    from samcli.commands.deploy.deploy_context import DeployContext
    from samcli.commands.package.package_context import PackageContext

LOG = logging.getLogger(__name__)

HELP_TEXT = """
Update/Sync local artifacts to AWS

By default, the sync command runs a full stack update. You can specify --code or --watch to switch modes.
\b
Sync also supports nested stacks and nested stack resources. For example

$ sam sync --code --stack-name {stack} --resource-id \\
{ChildStack}/{ResourceId}

Running --watch with --code option will provide a way to run code synchronization only, that will speed up start time
and will skip any template change. Please remember to update your deployed stack by running without --code option.

$ sam sync --code --watch --stack-name {stack} 

"""

SYNC_INFO_TEXT = """
The SAM CLI will use the AWS Lambda, Amazon API Gateway, and AWS StepFunctions APIs to upload your code without 
performing a CloudFormation deployment. This will cause drift in your CloudFormation stack. 
**The sync command should only be used against a development stack**.
"""

SYNC_CONFIRMATION_TEXT = """
Confirm that you are synchronizing a development stack.

Enter Y to proceed with the command, or enter N to cancel:
"""


SHORT_HELP = "Sync a project to AWS"

DEFAULT_TEMPLATE_NAME = "template.yaml"
DEFAULT_CAPABILITIES = ("CAPABILITY_NAMED_IAM", "CAPABILITY_AUTO_EXPAND")


@click.command("sync", help=HELP_TEXT, short_help=SHORT_HELP)
@configuration_option(provider=TomlProvider(section="parameters"))
@template_option_without_build
@click.option(
    "--code",
    is_flag=True,
    help="Sync code resources. This includes Lambda Functions, API Gateway, and Step Functions.",
    cls=ClickMutex,
)
@click.option(
    "--watch",
    is_flag=True,
    help="Watch local files and automatically sync with remote.",
    cls=ClickMutex,
)
@click.option(
    "--resource-id",
    multiple=True,
    help="Sync code for all the resources with the ID. To sync a resource within a nested stack, "
    "use the following pattern {ChildStack}/{logicalId}.",
)
@click.option(
    "--resource",
    multiple=True,
    type=click.Choice(SyncCodeResources.values(), case_sensitive=True),
    help=f"Sync code for all resources of the given resource type. Accepted values are {SyncCodeResources.values()}",
)
@click.option(
    "--dependency-layer/--no-dependency-layer",
    default=True,
    is_flag=True,
    help="This option separates the dependencies of individual function into another layer, for speeding up the sync."
    "process",
)
@stack_name_option(required=True)  # pylint: disable=E1120
@base_dir_option
@use_container_build_option
@image_repository_option
@image_repositories_option
@s3_bucket_option(disable_callback=True)  # pylint: disable=E1120
@s3_prefix_option
@kms_key_id_option
@role_arn_option
@parameter_override_option
@cli_framework_options
@aws_creds_options
@metadata_option
@notification_arns_option
@tags_option
@capabilities_option(default=DEFAULT_CAPABILITIES)  # pylint: disable=E1120
@pass_context
@track_command
@track_long_event("SyncUsed", "Start", "SyncUsed", "End")
@image_repository_validation
@track_template_warnings([CodeDeployWarning.__name__, CodeDeployConditionWarning.__name__])
@check_newer_version
@print_cmdline_args
@unsupported_command_cdk()
@command_exception_handler
def cli(
    ctx: Context,
    template_file: str,
    code: bool,
    watch: bool,
    resource_id: Optional[Tuple[str]],
    resource: Optional[Tuple[str]],
    dependency_layer: bool,
    stack_name: str,
    base_dir: Optional[str],
    parameter_overrides: dict,
    image_repository: str,
    image_repositories: Optional[Tuple[str]],
    s3_bucket: str,
    s3_prefix: str,
    kms_key_id: str,
    capabilities: Optional[List[str]],
    role_arn: Optional[str],
    notification_arns: Optional[List[str]],
    tags: dict,
    metadata: dict,
    use_container: bool,
    config_file: str,
    config_env: str,
) -> None:
    """
    `sam sync` command entry point
    """
    mode = _get_mode_value_from_envvar("SAM_BUILD_MODE", choices=["debug"])
    # All logic must be implemented in the ``do_cli`` method. This helps with easy unit testing

    do_cli(
        template_file,
        code,
        watch,
        resource_id,
        resource,
        dependency_layer,
        stack_name,
        ctx.region,
        ctx.profile,
        base_dir,
        parameter_overrides,
        mode,
        image_repository,
        image_repositories,
        s3_bucket,
        s3_prefix,
        kms_key_id,
        capabilities,
        role_arn,
        notification_arns,
        tags,
        metadata,
        use_container,
        config_file,
        config_env,
        None,  # TODO: replace with build_in_source once it's added as a click option
    )  # pragma: no cover


def do_cli(
    template_file: str,
    code: bool,
    watch: bool,
    resource_id: Optional[Tuple[str]],
    resource: Optional[Tuple[str]],
    dependency_layer: bool,
    stack_name: str,
    region: str,
    profile: str,
    base_dir: Optional[str],
    parameter_overrides: dict,
    mode: Optional[str],
    image_repository: str,
    image_repositories: Optional[Tuple[str]],
    s3_bucket: str,
    s3_prefix: str,
    kms_key_id: str,
    capabilities: Optional[List[str]],
    role_arn: Optional[str],
    notification_arns: Optional[List[str]],
    tags: dict,
    metadata: dict,
    use_container: bool,
    config_file: str,
    config_env: str,
    build_in_source: Optional[bool],
) -> None:
    """
    Implementation of the ``cli`` method
    """
    from samcli.cli.global_config import GlobalConfig
    from samcli.commands.build.build_context import BuildContext
    from samcli.commands.deploy.deploy_context import DeployContext
    from samcli.commands.package.package_context import PackageContext
    from samcli.lib.utils import osutils

    global_config = GlobalConfig()
    if not global_config.is_accelerate_opt_in_stack(template_file, stack_name):
        if not click.confirm(Colored().yellow(SYNC_INFO_TEXT + SYNC_CONFIRMATION_TEXT), default=True):
            return
        global_config.set_accelerate_opt_in_stack(template_file, stack_name)
    else:
        LOG.info(Colored().yellow(SYNC_INFO_TEXT))

    s3_bucket_name = s3_bucket or manage_stack(profile=profile, region=region)

    if dependency_layer is True:
        dependency_layer = check_enable_dependency_layer(template_file)

    # Note: ADL with use-container is not supported yet. Remove this logic once its supported.
    if use_container and dependency_layer:
        LOG.info(
            "Note: Automatic Dependency Layer is not yet supported with use-container. \
            sam sync will be run without Automatic Dependency Layer."
        )
        dependency_layer = False

    build_dir = DEFAULT_BUILD_DIR_WITH_AUTO_DEPENDENCY_LAYER if dependency_layer else DEFAULT_BUILD_DIR
    LOG.debug("Using build directory as %s", build_dir)
    EventTracker.track_event("UsedFeature", "Accelerate")

    with BuildContext(
        resource_identifier=None,
        template_file=template_file,
        base_dir=base_dir,
        build_dir=build_dir,
        cache_dir=DEFAULT_CACHE_DIR,
        clean=True,
        use_container=use_container,
        cached=True,
        parallel=True,
        parameter_overrides=parameter_overrides,
        mode=mode,
        create_auto_dependency_layer=dependency_layer,
        stack_name=stack_name,
        print_success_message=False,
        locate_layer_nested=True,
        build_in_source=build_in_source,
    ) as build_context:
        built_template = os.path.join(build_dir, DEFAULT_TEMPLATE_NAME)

        with osutils.tempfile_platform_independent() as output_template_file:
            with PackageContext(
                template_file=built_template,
                s3_bucket=s3_bucket_name,
                image_repository=image_repository,
                image_repositories=image_repositories,
                s3_prefix=s3_prefix,
                kms_key_id=kms_key_id,
                output_template_file=output_template_file.name,
                no_progressbar=True,
                metadata=metadata,
                region=region,
                profile=profile,
                use_json=False,
                force_upload=True,
            ) as package_context:

                # 500ms of sleep time between stack checks and describe stack events.
                DEFAULT_POLL_DELAY = 0.5
                try:
                    poll_delay = float(os.getenv("SAM_CLI_POLL_DELAY", str(DEFAULT_POLL_DELAY)))
                except ValueError:
                    poll_delay = DEFAULT_POLL_DELAY
                if poll_delay <= 0:
                    poll_delay = DEFAULT_POLL_DELAY

                with DeployContext(
                    template_file=output_template_file.name,
                    stack_name=stack_name,
                    s3_bucket=s3_bucket_name,
                    image_repository=image_repository,
                    image_repositories=image_repositories,
                    no_progressbar=True,
                    s3_prefix=s3_prefix,
                    kms_key_id=kms_key_id,
                    parameter_overrides=parameter_overrides,
                    capabilities=capabilities,
                    role_arn=role_arn,
                    notification_arns=notification_arns,
                    tags=tags,
                    region=region,
                    profile=profile,
                    no_execute_changeset=True,
                    fail_on_empty_changeset=True,
                    confirm_changeset=False,
                    use_changeset=False,
                    force_upload=True,
                    signing_profiles=None,
                    disable_rollback=False,
                    poll_delay=poll_delay,
                    on_failure=None,
                ) as deploy_context:
                    with SyncContext(
                        dependency_layer, build_context.build_dir, build_context.cache_dir
                    ) as sync_context:
                        if watch:
                            execute_watch(
                                template_file,
                                build_context,
                                package_context,
                                deploy_context,
                                sync_context,
                                dependency_layer,
                                code,
                            )
                        elif code:
                            execute_code_sync(
                                template_file,
                                build_context,
                                deploy_context,
                                sync_context,
                                resource_id,
                                resource,
                                dependency_layer,
                            )
                        else:
                            execute_infra_contexts(build_context, package_context, deploy_context)


def execute_infra_contexts(
    build_context: "BuildContext",
    package_context: "PackageContext",
    deploy_context: "DeployContext",
) -> None:
    """Executes the sync for infra.

    Parameters
    ----------
    build_context : BuildContext
        BuildContext
    package_context : PackageContext
        PackageContext
    deploy_context : DeployContext
        DeployContext
    """
    LOG.debug("Executing the build using build context.")
    build_context.run()
    LOG.debug("Executing the packaging using package context.")
    package_context.run()
    LOG.debug("Executing the deployment using deploy context.")
    deploy_context.run()


def execute_code_sync(
    template: str,
    build_context: "BuildContext",
    deploy_context: "DeployContext",
    sync_context: "SyncContext",
    resource_ids: Optional[Tuple[str]],
    resource_types: Optional[Tuple[str]],
    auto_dependency_layer: bool,
) -> None:
    """Executes the sync flow for code.

    Parameters
    ----------
    template : str
        Template file name
    build_context : BuildContext
        BuildContext
    deploy_context : DeployContext
        DeployContext
    sync_context: SyncContext
        SyncContext object that obtains sync information.
    resource_ids : List[str]
        List of resource IDs to be synced.
    resource_types : List[str]
        List of resource types to be synced.
    auto_dependency_layer: bool
        Boolean flag to whether enable certain sync flows for auto dependency layer feature
    """
    stacks = SamLocalStackProvider.get_stacks(template)[0]
    factory = SyncFlowFactory(build_context, deploy_context, sync_context, stacks, auto_dependency_layer)
    factory.load_physical_id_mapping()
    executor = SyncFlowExecutor()

    sync_flow_resource_ids: Set[ResourceIdentifier] = (
        get_unique_resource_ids(stacks, resource_ids, resource_types)
        if resource_ids or resource_types
        else set(get_all_resource_ids(stacks))
    )

    for resource_id in sync_flow_resource_ids:
        sync_flow = factory.create_sync_flow(resource_id)
        if sync_flow:
            executor.add_sync_flow(sync_flow)
        else:
            LOG.warning("Cannot create SyncFlow for %s. Skipping.", resource_id)
    executor.execute()


def execute_watch(
    template: str,
    build_context: "BuildContext",
    package_context: "PackageContext",
    deploy_context: "DeployContext",
    sync_context: "SyncContext",
    auto_dependency_layer: bool,
    skip_infra_syncs: bool,
):
    """Start sync watch execution

    Parameters
    ----------
    template : str
        Template file path
    build_context : BuildContext
        BuildContext
    package_context : PackageContext
        PackageContext
    deploy_context : DeployContext
        DeployContext
    sync_context: SyncContext
        SyncContext object that obtains sync information.
    auto_dependency_layer: bool
        Boolean flag to whether enable certain sync flows for auto dependency layer feature.
    skip_infra_syncs: bool
        Boolean flag to determine if only ececute code syncs.
    """
    watch_manager = WatchManager(
        template, build_context, package_context, deploy_context, sync_context, auto_dependency_layer, skip_infra_syncs
    )
    watch_manager.start()


def check_enable_dependency_layer(template_file: str):
    """
    Check if auto dependency layer should be enabled
    :param template_file: template file string
    :return: True if ADL should be enabled, False otherwise
    """
    stacks, _ = SamLocalStackProvider.get_stacks(template_file)
    for stack in stacks:
        esbuild = EsbuildBundlerManager(stack)
        if esbuild.esbuild_configured():
            # Disable ADL if esbuild is configured. esbuild already makes the package size
            # small enough to ensure that ADL isn't needed to improve performance
            click.secho("esbuild is configured, disabling auto dependency layer.", fg="yellow")
            return False
    return True



"""
Custom exception used by Local Lambda execution
"""


class FunctionNotFound(Exception):
    """
    Raised when the requested Lambda function is not found
    """


class ResourceNotFound(Exception):
    """
    Raised when the requested resource is not found
    """



"""
Class that provides all nested stacks from a given SAM template
"""
import logging
import os
from typing import Dict, Iterator, List, Optional, Tuple, Union, cast
from urllib.parse import unquote, urlparse

from samcli.commands._utils.template import TemplateNotFoundException, get_template_data
from samcli.lib.providers.exceptions import RemoteStackLocationNotSupported
from samcli.lib.providers.provider import Stack, get_full_path
from samcli.lib.providers.sam_base_provider import SamBaseProvider
from samcli.lib.utils.resources import AWS_CLOUDFORMATION_STACK, AWS_SERVERLESS_APPLICATION

LOG = logging.getLogger(__name__)


class SamLocalStackProvider(SamBaseProvider):
    """
    Fetches and returns local nested stacks from a SAM Template. The SAM template passed to this provider is assumed
    to be valid, normalized and a dictionary.
    It may or may not contain a stack.
    """

    def __init__(
        self,
        template_file: str,
        stack_path: str,
        template_dict: Dict,
        parameter_overrides: Optional[Dict] = None,
        global_parameter_overrides: Optional[Dict] = None,
    ):
        """
        Initialize the class with SAM template data. The SAM template passed to this provider is assumed
        to be valid and a dictionary. This class will perform template normalization to remove structures
        like ``Globals``, resolve intrinsic functions etc.
        This class does not perform any syntactic validation of the template.
        After the class is initialized, any changes to the ``template_dict`` will not be reflected in here.
        You need to explicitly update the class with new template, if necessary.
        :param str template_file: SAM Stack Template file path
        :param str stack_path: SAM Stack stack_path (See samcli.lib.providers.provider.Stack.stack_path)
        :param dict template_dict: SAM Template as a dictionary
        :param dict parameter_overrides: Optional dictionary of values for SAM template parameters that might want
            to get substituted within the template
        :param dict global_parameter_overrides: Optional dictionary of values for SAM template global parameters that
            might want to get substituted within the template and all its child templates
        """

        self._template_file = template_file
        self._stack_path = stack_path
        self._template_dict = self.get_template(
            template_dict,
            SamLocalStackProvider.merge_parameter_overrides(parameter_overrides, global_parameter_overrides),
        )
        self._resources = self._template_dict.get("Resources", {})
        self._global_parameter_overrides = global_parameter_overrides

        # Store a map of stack name to stack information for quick reference -> self._stacks
        # and detect remote stacks -> self._remote_stack_full_paths
        self._stacks: Dict[str, Stack] = {}
        self.remote_stack_full_paths: List[str] = []
        self._extract_stacks()

        LOG.debug("%d stacks found in the template", len(self._stacks))

    def get(self, name: str) -> Optional[Stack]:
        """
        Returns the application given name or LogicalId of the application.
        Every SAM resource has a logicalId, but it may
        also have a application name. This method searches only for LogicalID and returns the application that matches
        it.
        :param string name: Name of the application
        :return Function: namedtuple containing the Application information if application is found.
                          None, if application is not found
        :raises ValueError If name is not given
        """
        for f in self.get_all():
            if f.name == name:
                return f

        return None

    def get_all(self) -> Iterator[Stack]:
        """
        Yields all the applications available in the SAM Template.
        :yields Application: map containing the application information
        """

        for _, stack in self._stacks.items():
            yield stack

    def _extract_stacks(self) -> None:
        """
        Extracts and returns nested application information from the given dictionary of SAM/CloudFormation resources.
        This method supports applications defined with AWS::Serverless::Application
        The dictionary of application LogicalId to the Application object will be assigned to self._stacks.
        If child stacks with remote URL are detected, their full paths are recorded in self._remote_stack_full_paths.
        """

        for name, resource in self._resources.items():

            resource_type = resource.get("Type")
            resource_properties = resource.get("Properties", {})
            resource_metadata = resource.get("Metadata", None)
            # Add extra metadata information to properties under a separate field.
            if resource_metadata:
                resource_properties["Metadata"] = resource_metadata

            stack: Optional[Stack] = None
            try:
                if resource_type == AWS_SERVERLESS_APPLICATION:
                    stack = SamLocalStackProvider._convert_sam_application_resource(
                        self._template_file, self._stack_path, name, resource_properties
                    )
                if resource_type == AWS_CLOUDFORMATION_STACK:
                    stack = SamLocalStackProvider._convert_cfn_stack_resource(
                        self._template_file, self._stack_path, name, resource_properties
                    )
            except RemoteStackLocationNotSupported:
                self.remote_stack_full_paths.append(get_full_path(self._stack_path, name))

            if stack:
                self._stacks[name] = stack

            # We don't care about other resource types. Just ignore them

    @staticmethod
    def _convert_sam_application_resource(
        template_file: str,
        stack_path: str,
        name: str,
        resource_properties: Dict,
        global_parameter_overrides: Optional[Dict] = None,
    ) -> Optional[Stack]:
        location = resource_properties.get("Location")

        if isinstance(location, dict):
            raise RemoteStackLocationNotSupported()

        location = cast(str, location)
        if SamLocalStackProvider.is_remote_url(location):
            raise RemoteStackLocationNotSupported()
        if location.startswith("file://"):
            location = unquote(urlparse(location).path)
        else:
            location = SamLocalStackProvider.normalize_resource_path(template_file, location)

        return Stack(
            parent_stack_path=stack_path,
            name=name,
            location=location,
            parameters=SamLocalStackProvider.merge_parameter_overrides(
                resource_properties.get("Parameters", {}), global_parameter_overrides
            ),
            template_dict=get_template_data(location),
            metadata=resource_properties.get("Metadata", {}),
        )

    @staticmethod
    def _convert_cfn_stack_resource(
        template_file: str,
        stack_path: str,
        name: str,
        resource_properties: Dict,
        global_parameter_overrides: Optional[Dict] = None,
    ) -> Optional[Stack]:
        template_url = resource_properties.get("TemplateURL")

        if isinstance(template_url, dict):
            # This happens when TemplateURL has unresolvable intrinsic functions
            # and it usually happens in CDK generated template files (#2832).
            raise RemoteStackLocationNotSupported()

        template_url = cast(str, template_url)
        if SamLocalStackProvider.is_remote_url(template_url):
            raise RemoteStackLocationNotSupported()
        if template_url.startswith("file://"):
            template_url = unquote(urlparse(template_url).path)
        else:
            template_url = SamLocalStackProvider.normalize_resource_path(template_file, template_url)

        return Stack(
            parent_stack_path=stack_path,
            name=name,
            location=template_url,
            parameters=SamLocalStackProvider.merge_parameter_overrides(
                resource_properties.get("Parameters", {}), global_parameter_overrides
            ),
            template_dict=get_template_data(template_url),
            metadata=resource_properties.get("Metadata", {}),
        )

    @staticmethod
    def get_stacks(
        template_file: Optional[str] = None,
        stack_path: str = "",
        name: str = "",
        parameter_overrides: Optional[Dict] = None,
        global_parameter_overrides: Optional[Dict] = None,
        metadata: Optional[Dict] = None,
        template_dictionary: Optional[Dict] = None,
    ) -> Tuple[List[Stack], List[str]]:
        """
        Recursively extract stacks from a template file.

        Parameters
        ----------
        template_file: str
            the file path of the template to extract stacks from. Only one of either template_dict or template_file
            is required
        stack_path: str
            the stack path of the parent stack, for root stack, it is ""
        name: str
            the name of the stack associated with the template_file, for root stack, it is ""
        parameter_overrides: Optional[Dict]
            Optional dictionary of values for SAM template parameters that might want
            to get substituted within the template
        global_parameter_overrides: Optional[Dict]
            Optional dictionary of values for SAM template global parameters
            that might want to get substituted within the template and its child templates
        metadata: Optional[Dict]
            Optional dictionary of nested stack resource metadata values.
        template_dictionary: Optional[Dict]
            dictionary representing the sam template. Only one of either template_dict or template_file is required

        Returns
        -------
        stacks: List[Stack]
            The list of stacks extracted from template_file
        remote_stack_full_paths : List[str]
            The list of full paths of detected remote stacks
        """
        template_dict: dict
        if template_file:
            template_dict = get_template_data(template_file)
        elif template_dictionary:
            template_file = ""
            template_dict = template_dictionary
        else:
            raise TemplateNotFoundException(
                message="A template file or a template dict is required but both are missing."
            )

        stacks = [
            Stack(
                stack_path,
                name,
                template_file,
                SamLocalStackProvider.merge_parameter_overrides(parameter_overrides, global_parameter_overrides),
                template_dict,
                metadata,
            )
        ]
        remote_stack_full_paths: List[str] = []

        current = SamLocalStackProvider(
            template_file, stack_path, template_dict, parameter_overrides, global_parameter_overrides
        )
        remote_stack_full_paths.extend(current.remote_stack_full_paths)

        for child_stack in current.get_all():
            stacks_in_child, remote_stack_full_paths_in_child = SamLocalStackProvider.get_stacks(
                child_stack.location,
                os.path.join(stack_path, stacks[0].stack_id),
                child_stack.name,
                child_stack.parameters,
                global_parameter_overrides,
                child_stack.metadata,
            )
            stacks.extend(stacks_in_child)
            remote_stack_full_paths.extend(remote_stack_full_paths_in_child)

        return stacks, remote_stack_full_paths

    @staticmethod
    def is_remote_url(url: str) -> bool:
        return any([url.startswith(prefix) for prefix in ["s3://", "http://", "https://"]])

    @staticmethod
    def find_root_stack(stacks: List[Stack]) -> Stack:
        candidates = [stack for stack in stacks if stack.is_root_stack]
        if not candidates:
            stacks_str = ", ".join([stack.stack_path for stack in stacks])
            raise ValueError(f"{stacks_str} does not contain a root stack")
        return candidates[0]

    @staticmethod
    def merge_parameter_overrides(
        parameter_overrides: Optional[Dict], global_parameter_overrides: Optional[Dict]
    ) -> Dict:
        """
        Combine global parameters and stack-specific parameters.
        Right now the only global parameter override available is AWS::Region (via --region in "sam local"),
        and AWS::Region won't appear in normal stack-specific parameter_overrides, so we don't
        specify which type of parameters have high precedence.

        Parameters
        ----------
        parameter_overrides: Optional[Dict]
            stack-specific parameters
        global_parameter_overrides: Optional[Dict]
            global parameters

        Returns
        -------
        Dict
            merged dict containing both global and stack-specific parameters
        """
        merged_parameter_overrides = {}
        merged_parameter_overrides.update(global_parameter_overrides or {})
        merged_parameter_overrides.update(parameter_overrides or {})
        return merged_parameter_overrides

    @staticmethod
    def normalize_resource_path(stack_file_path: str, path: str) -> str:
        """
        Convert resource paths found in nested stack to ones resolvable from root stack.
        For example,
            root stack                -> template.yaml
            child stack               -> folder/template.yaml
            a resource in child stack -> folder/resource
        the resource path is "resource" because it is extracted from child stack, the path is relative to child stack.
        here we normalize the resource path into relative paths to root stack, which is "folder/resource"

        * since stack_file_path might be a symlink, os.path.join() won't be able to derive the correct path.
          for example, stack_file_path = 'folder/t.yaml' -> '../folder2/t.yaml' and the path = 'src'
          the correct normalized path being returned should be '../folder2/t.yaml' but if we don't resolve the
          symlink first, it would return 'folder/src.'

        * symlinks on Windows might not work properly.
          https://stackoverflow.com/questions/43333640/python-os-path-realpath-for-symlink-in-windows
          For example, using Python 3.7, realpath() is a no-op (same as abspath):
            ```
            Python 3.7.8 (tags/v3.7.8:4b47a5b6ba, Jun 28 2020, 08:53:46) [MSC v.1916 64 bit (AMD64)] on win32
            Type "help", "copyright", "credits" or "license" for more information.
            >>> import os
            >>> os.symlink('some\\path', 'link1')
            >>> os.path.realpath('link1')
            'C:\\Users\\Administrator\\AppData\\Local\\Programs\\Python\\Python37\\link1'
            >>> os.path.islink('link1')
            True
            ```
          For Python 3.8, according to manual tests, 3.8.8 can resolve symlinks correctly while 3.8.0 cannot.


        Parameters
        ----------
        stack_file_path
            The file path of the stack containing the resource
        path
            the raw path read from the template dict

        Returns
        -------
        str
            the normalized path relative to root stack

        """
        if os.path.isabs(path):
            return path

        if os.path.islink(stack_file_path):
            # os.path.realpath() always returns an absolute path while
            # the return value of this method will show up in build artifacts,
            # in case customers move the build artifacts among different machines (e.g., git or file sharing)
            # absolute paths are not robust as relative paths. So here prefer to use relative path.
            stack_file_path = os.path.relpath(os.path.realpath(stack_file_path))

        return os.path.normpath(os.path.join(os.path.dirname(stack_file_path), path))


def is_local_path(path: Union[Dict, str]) -> bool:
    return bool(path) and not isinstance(path, dict) and not SamLocalStackProvider.is_remote_url(path)


def get_local_path(path: str, parent_path: str) -> str:
    if path.startswith("file://"):
        path = unquote(urlparse(path).path)
    else:
        path = SamLocalStackProvider.normalize_resource_path(parent_path, path)
    return path



"""Base SyncFlow for StepFunctions"""
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from samcli.lib.providers.exceptions import MissingLocalDefinition
from samcli.lib.providers.provider import ResourceIdentifier, Stack, get_resource_by_id
from samcli.lib.sync.exceptions import InfraSyncRequiredError
from samcli.lib.sync.sync_flow import ResourceAPICall, SyncFlow, get_definition_path
from samcli.lib.utils.hash import str_checksum

if TYPE_CHECKING:  # pragma: no cover
    from samcli.commands.build.build_context import BuildContext
    from samcli.commands.deploy.deploy_context import DeployContext
    from samcli.commands.sync.sync_context import SyncContext

LOG = logging.getLogger(__name__)


class StepFunctionsSyncFlow(SyncFlow):
    _state_machine_identifier: str
    _stepfunctions_client: Any
    _stacks: List[Stack]
    _definition_uri: Optional[Path]
    _states_definition: Optional[str]

    def __init__(
        self,
        state_machine_identifier: str,
        build_context: "BuildContext",
        deploy_context: "DeployContext",
        sync_context: "SyncContext",
        physical_id_mapping: Dict[str, str],
        stacks: List[Stack],
    ):
        """
        Parameters
        ----------
        state_machine_identifier : str
            State Machine resource identifier that need to be synced.
        build_context : BuildContext
            BuildContext used for build related parameters
        deploy_context : BuildContext
            DeployContext used for this deploy related parameters
        sync_context: SyncContext
            SyncContext object that obtains sync information.
        physical_id_mapping : Dict[str, str]
            Mapping between resource logical identifier and physical identifier
        stacks : List[Stack], optional
            List of stacks containing a root stack and optional nested stacks
        """
        super().__init__(
            build_context,
            deploy_context,
            sync_context,
            physical_id_mapping,
            log_name="StepFunctions " + state_machine_identifier,
            stacks=stacks,
        )
        self._state_machine_identifier = state_machine_identifier
        self._resource = get_resource_by_id(self._stacks, ResourceIdentifier(self._state_machine_identifier))
        self._stepfunctions_client = None
        self._definition_uri = None
        self._states_definition = None

    @property
    def sync_state_identifier(self) -> str:
        """
        Sync state is the unique identifier for each sync flow
        In sync state toml file we will store
        Key as StepFunctionsSyncFlow:StepFunctionsLogicalId
        Value as state machine definition hash
        """
        return self.__class__.__name__ + ":" + self._state_machine_identifier

    def set_up(self) -> None:
        super().set_up()
        self._stepfunctions_client = self._boto_client("stepfunctions")

    def gather_resources(self) -> None:
        if not self._resource:
            return
        definition_substitutions = self._resource.get("Properties", dict()).get("DefinitionSubstitutions", None)
        if definition_substitutions:
            raise InfraSyncRequiredError(self._state_machine_identifier, "DefinitionSubstitutions field is specified.")
        self._definition_uri = self._get_definition_file(self._state_machine_identifier)
        self._states_definition = self._process_definition_file()
        if self._states_definition:
            self._local_sha = str_checksum(self._states_definition)

    def _process_definition_file(self) -> Optional[str]:
        if self._definition_uri is None:
            return None
        with open(str(self._definition_uri), "r", encoding="utf-8") as states_file:
            states_data = states_file.read()
            return states_data

    def _get_definition_file(self, state_machine_identifier: str) -> Optional[Path]:
        if not self._resource:
            return None
        return get_definition_path(
            self._resource,
            state_machine_identifier,
            self._build_context.use_base_dir,
            self._build_context.base_dir,
            self._stacks,
        )

    def compare_remote(self) -> bool:
        # Not comparing with remote right now, instead only making update api calls
        # Note: describe state machine has a better rate limit then update state machine
        # So if we face any throttling issues, comparing should be desired
        return False

    def gather_dependencies(self) -> List[SyncFlow]:
        return []

    def _get_resource_api_calls(self) -> List[ResourceAPICall]:
        return []

    def _equality_keys(self):
        return self._state_machine_identifier

    def sync(self) -> None:
        state_machine_arn = self.get_physical_id(self._state_machine_identifier)
        if self._definition_uri is None:
            raise MissingLocalDefinition(ResourceIdentifier(self._state_machine_identifier), "DefinitionUri")
        LOG.debug("%sTrying to update State Machine definition", self.log_prefix)
        response = self._stepfunctions_client.update_state_machine(
            stateMachineArn=state_machine_arn, definition=self._states_definition
        )
        LOG.debug("%sUpdate State Machine: %s", self.log_prefix, response)



"""
IaC Exceptions
"""
from typing import Optional

from samcli.commands.exceptions import UserException


class InvalidIaCPluginException(UserException):
    def __init__(self, files: Optional[list] = None):
        if files is None:
            files = []
        msg = "Could not determine the plugin type from the provided files:\n\n{files}"
        UserException.__init__(self, msg.format(files=", ".join(files)))


class InvalidProjectTypeException(UserException):
    def __init__(self, msg):
        UserException.__init__(self, msg)



"""
Provides all Warnings checkers for sam template
"""
import logging
from typing import Dict

LOG = logging.getLogger(__name__)


class TemplateWarning:
    """
    Top level class which all warnings should extend from.
    """

    def check(self, template_dict):  # pylint: disable=no-self-use
        raise Exception("NotImplementedException")


def _get_deployment_preferences_status(function):
    """
    Takes a AWS::Serverless::Function resource and checks if resource have a deployment preferences applied
    to it. If DeploymentPreference found then it returns its status if it is enabled or not.
    """
    deployment_preference = function.get("Properties", {}).get("DeploymentPreference", None)
    if not deployment_preference:
        # Missing deployment preferences treated as not enabled.
        return False
    return deployment_preference.get("Enabled", True)  # enabled by default


class TemplateWarningsChecker:
    def __init__(self):
        self.all_warnings = {
            CodeDeployWarning.__name__: CodeDeployWarning(),
            CodeDeployConditionWarning.__name__: CodeDeployConditionWarning(),
        }

    def check_template_for_warning(self, warning_name, template_dict):
        """
        Checks provided template against the warning based on warning_name.
        Parameters
        ----------
        warning_name: Name of warning which needs to be checked.
        template_dict: template dict

        Returns
        -------
        warning_message if warning detected. None if no warning found.
        """
        if not template_dict:
            return None
        warning = self.all_warnings.get(warning_name, None)
        if not warning:
            LOG.error("UnknownWarning name found: %s", warning_name)
            return None

        should_warn, warning_message = warning.check(template_dict)
        if should_warn:
            return warning_message
        return None


class CodeDeployWarning(TemplateWarning):
    WARNING_MESSAGE = """\
Your template includes a deployment configuration that will create an additional unecessary CodeDeploy Service Role.
By 9/25 the SAM service will no longer create this extra role, and any applications deployed after this date that 
directly reference this role will produce an error as the role will be removed. For more information on this issue 
and how to mitigate it, please read these docs[1]

[1] https://github.com/aws/aws-sam-cli/wiki/08-2020-codeploy-servicerole
    """

    def check(self, template_dict):
        """
        Checking if template dictionary have CodeDeployWarning or not.
        """
        functions = [
            resource
            for (_, resource) in template_dict.get("Resources", {}).items()
            if resource.get("Type", "") == "AWS::Serverless::Function"
        ]
        deployment_features_enabled_count = sum(
            1 for function in functions if _get_deployment_preferences_status(function)
        )
        deployment_features_disabled_count = sum(
            1 for function in functions if not _get_deployment_preferences_status(function)
        )

        send_warning = deployment_features_enabled_count > 0 and deployment_features_disabled_count > 0
        return (send_warning, self.WARNING_MESSAGE) if send_warning else (send_warning, "")


class CodeDeployConditionWarning(TemplateWarning):
    WARNING_MESSAGE = """\
Your template includes a deployment configuration with a Condition attached to it. SAM currently has a bug that
ignores conditions for DeploymentPreference, causing CodeDeploy DeploymentGroups to be created in error.
After October 23, 2020 the SAM service will fix this bug, causing subsequent deployments to remove these CodeDeploy 
DeploymentGroups if the attached Condition is false. For more information on this issue and how to mitigate it, 
please read these docs[1]

[1] https://github.com/aws/aws-sam-cli/wiki/08-2020-codeploy-deploymentgroup-condition
    """

    def check(self, template_dict):
        """
        Checking if template dictionary have Function with Condition and DeploymentPreferences which
        will trigger this warning.
        """
        functions = [
            resource
            for (_, resource) in template_dict.get("Resources", {}).items()
            if resource.get("Type", "") == "AWS::Serverless::Function"
        ]
        for function in functions:
            if self._have_condition(function) and self._have_deployment_preferences(function):
                return (True, self.WARNING_MESSAGE)
        return (False, "")

    @staticmethod
    def _have_condition(function: Dict) -> bool:
        condition = function.get("Condition", None)
        return condition is not None

    @staticmethod
    def _have_deployment_preferences(function: Dict) -> bool:
        deployment_preference = function.get("Properties", {}).get("DeploymentPreference", None)
        return deployment_preference is not None



"""
Terraform prepare property builder
"""
from typing import Any, Dict, Optional

from samcli.hook_packages.terraform.hooks.prepare.resource_linking import _resolve_resource_attribute
from samcli.hook_packages.terraform.hooks.prepare.types import (
    PropertyBuilder,
    PropertyBuilderMapping,
    ResourceTranslator,
    TFResource,
)
from samcli.lib.hook.exceptions import PrepareHookException
from samcli.lib.utils.packagetype import IMAGE, ZIP
from samcli.lib.utils.resources import AWS_LAMBDA_FUNCTION as CFN_AWS_LAMBDA_FUNCTION
from samcli.lib.utils.resources import AWS_LAMBDA_LAYERVERSION as CFN_AWS_LAMBDA_LAYER_VERSION

REMOTE_DUMMY_VALUE = "<<REMOTE DUMMY VALUE - RAISE ERROR IF IT IS STILL THERE>>"
TF_AWS_LAMBDA_FUNCTION = "aws_lambda_function"
TF_AWS_LAMBDA_LAYER_VERSION = "aws_lambda_layer_version"


def _build_code_property(tf_properties: dict, resource: TFResource) -> Any:
    """
    Builds the Code property of a CloudFormation AWS Lambda Function out of the
    properties of the equivalent terraform resource

    Parameters
    ----------
    tf_properties: dict
        Properties of the terraform AWS Lambda function resource
    resource: TFResource
        Configuration terraform resource

    Returns
    -------
    dict
        The built Code property of a CloudFormation AWS Lambda Function resource
    """
    filename = tf_properties.get("filename")
    if filename:
        return filename

    code = {}
    tf_cfn_prop_names = [
        ("s3_bucket", "S3Bucket"),
        ("s3_key", "S3Key"),
        ("image_uri", "ImageUri"),
        ("s3_object_version", "S3ObjectVersion"),
    ]
    for tf_prop_name, cfn_prop_name in tf_cfn_prop_names:
        tf_prop_value = tf_properties.get(tf_prop_name)
        if tf_prop_value is not None:
            code[cfn_prop_name] = tf_prop_value

    package_type = tf_properties.get("package_type", ZIP)

    # Get the S3 Bucket details from configuration in case if the customer is creating the S3 bucket in the tf project
    if package_type == ZIP and ("S3Bucket" not in code or "S3Key" not in code or "S3ObjectVersion" not in code):
        s3_bucket_tf_config_value = _resolve_resource_attribute(resource, "s3_bucket")
        s3_key_tf_config_value = _resolve_resource_attribute(resource, "s3_key")
        s3_object_version_tf_config_value = _resolve_resource_attribute(resource, "s3_object_version")
        if "S3Bucket" not in code and s3_bucket_tf_config_value:
            code["S3Bucket"] = REMOTE_DUMMY_VALUE
            code["S3Bucket_config_value"] = s3_bucket_tf_config_value
        if "S3Key" not in code and s3_key_tf_config_value:
            code["S3Key"] = REMOTE_DUMMY_VALUE
            code["S3Key_config_value"] = s3_key_tf_config_value
        if "S3ObjectVersion" not in code and s3_object_version_tf_config_value:
            code["S3ObjectVersion"] = REMOTE_DUMMY_VALUE
            code["S3ObjectVersion_config_value"] = s3_object_version_tf_config_value

    # Get the Image URI details from configuration in case if the customer is creating the ecr repo in the tf project
    if package_type == IMAGE and "ImageUri" not in code:
        image_uri_tf_config_value = _resolve_resource_attribute(resource, "image_uri")
        if image_uri_tf_config_value:
            code["ImageUri"] = REMOTE_DUMMY_VALUE

    return code


def _get_property_extractor(property_name: str) -> PropertyBuilder:
    """
    Returns a PropertyBuilder function to extract the given property from a dict

    Parameters
    ----------
    property_name: str
        The name of the property to extract

    Returns
    -------
    PropertyBuilder
        function that takes in a dict and extracts the given property name from it
    """
    return lambda properties, _: properties.get(property_name)


def _build_lambda_function_environment_property(tf_properties: dict, resource: TFResource) -> Optional[dict]:
    """
    Builds the Environment property of a CloudFormation AWS Lambda Function out of the
    properties of the equivalent terraform resource

    Parameters
    ----------
    tf_properties: dict
        Properties of the terraform AWS Lambda function resource
    resource: TFResource
        Configuration terraform resource

    Returns
    -------
    dict
        The built Environment property of a CloudFormation AWS Lambda Function resource
    """
    environment = tf_properties.get("environment")
    if not environment:
        return None

    for env in environment:
        variables = env.get("variables")
        if variables:
            return {"Variables": variables}

    # no variables
    return None


def _build_lambda_function_image_config_property(tf_properties: dict, resource: TFResource) -> Optional[dict]:
    """
    Builds the ImageConfig property of a CloudFormation AWS Lambda Function out of the
    properties of the equivalent terraform resource

    Parameters
    ----------
    tf_properties: dict
        Properties of the terraform AWS Lambda function resource
    resource: TFResource
        Configuration terraform resource

    Returns
    -------
    dict
        The built ImageConfig property of a CloudFormation AWS Lambda Function resource
    """
    image_config = tf_properties.get("image_config")
    if not image_config:
        return None

    _check_image_config_value(image_config)
    image_config = image_config[0]

    cfn_image_config = {}
    tf_cfn_prop_names = [
        ("command", "Command"),
        ("entry_point", "EntryPoint"),
        ("working_directory", "WorkingDirectory"),
    ]

    for tf_prop_name, cfn_prop_name in tf_cfn_prop_names:
        tf_prop_value = image_config.get(tf_prop_name)
        if tf_prop_value is not None:
            cfn_image_config[cfn_prop_name] = tf_prop_value

    return cfn_image_config


def _check_image_config_value(image_config: Any) -> bool:
    """
    validate if the image_config property value is as SAM CLI expects. If it is not valid, it will raise a
    PrepareHookException.

     Parameters
    ----------
    image_config: Any
        The aws_lambda resource's Image_config property value as read from the terraform plan output.

    Returns
    -------
    bool
        return True, if the image_config value as expects, and raise PrepareHookException if not as expected.
    """
    if not isinstance(image_config, list):
        raise PrepareHookException(
            f"AWS SAM CLI expects that the value of image_config of aws_lambda_function resource in "
            f"the terraform plan output to be of type list instead of {type(image_config)}"
        )
    if len(image_config) > 1:
        raise PrepareHookException(
            f"AWS SAM CLI expects that there is only one item in the  image_config property of "
            f"aws_lambda_function resource in the terraform plan output, but there are "
            f"{len(image_config)} items"
        )
    return True


AWS_LAMBDA_FUNCTION_PROPERTY_BUILDER_MAPPING: PropertyBuilderMapping = {
    "FunctionName": _get_property_extractor("function_name"),
    "Architectures": _get_property_extractor("architectures"),
    "Environment": _build_lambda_function_environment_property,
    "Code": _build_code_property,
    "Handler": _get_property_extractor("handler"),
    "PackageType": _get_property_extractor("package_type"),
    "Runtime": _get_property_extractor("runtime"),
    "Layers": _get_property_extractor("layers"),
    "Timeout": _get_property_extractor("timeout"),
    "ImageConfig": _build_lambda_function_image_config_property,
}

AWS_LAMBDA_LAYER_VERSION_PROPERTY_BUILDER_MAPPING: PropertyBuilderMapping = {
    "LayerName": _get_property_extractor("layer_name"),
    "CompatibleRuntimes": _get_property_extractor("compatible_runtimes"),
    "CompatibleArchitectures": _get_property_extractor("compatible_architectures"),
    "Content": _build_code_property,
}

RESOURCE_TRANSLATOR_MAPPING: Dict[str, ResourceTranslator] = {
    TF_AWS_LAMBDA_FUNCTION: ResourceTranslator(CFN_AWS_LAMBDA_FUNCTION, AWS_LAMBDA_FUNCTION_PROPERTY_BUILDER_MAPPING),
    TF_AWS_LAMBDA_LAYER_VERSION: ResourceTranslator(
        CFN_AWS_LAMBDA_LAYER_VERSION, AWS_LAMBDA_LAYER_VERSION_PROPERTY_BUILDER_MAPPING
    ),
}



""" Read info from runtime_config.json file"""

import json
import logging
from pathlib import Path

from click import ClickException

LOG = logging.getLogger(__name__)

CONFIG_FILE = Path(Path(__file__).resolve().parents[2], "runtime_config.json")
config = json.loads(CONFIG_FILE.read_text())


def get_app_template_repo_commit():
    """
    Returns
    -------
    the value of app_template_repo_commit

    """
    commit_hash = config.get("app_template_repo_commit", None)
    if not commit_hash:
        raise ClickException(
            message="Error when retrieving app_template_repo_commit, runtime_config.json file maybe invalid"
        )
    return commit_hash



"""
Implementation of the endpoints to table mapper
"""
from collections import OrderedDict
from typing import Any, Dict

from samcli.lib.list.list_interfaces import Mapper

NO_DATA = "-"
SPACING = ""
CLOUD_ENDPOINT = "CloudEndpoint"
METHODS = "Methods"


class EndpointsToTableMapper(Mapper):
    """
    Mapper class for mapping endpoints data for table output
    """

    def map(self, data: list) -> Dict[Any, Any]:
        """
        Maps data to the format needed for consumption by the table consumer

        Parameters
        ----------
        data: list
            List of dictionaries containing the entries of the endpoints data

        Returns
        -------
        table_data: Dict[Any, Any]
            Dictionary containing the information and data needed for the table consumer
            to output the data in table format
        """
        entry_list = []

        # Parse through the data object and separate out each data point we want to display.
        # If the data is none, default to using a "-"
        for endpoint in data:
            cloud_endpoint_furl_string = endpoint.get(CLOUD_ENDPOINT, NO_DATA)
            methods_string = NO_DATA
            cloud_endpoint_furl_multi_list = []

            # Build row of cloud endpoint data
            if isinstance(endpoint.get(CLOUD_ENDPOINT, NO_DATA), list) and endpoint.get(CLOUD_ENDPOINT, []):
                cloud_endpoint_furl_string = endpoint.get(CLOUD_ENDPOINT, [NO_DATA])[0]
                if len(endpoint.get(CLOUD_ENDPOINT, [])) > 1:
                    cloud_endpoint_furl_multi_list = endpoint.get(CLOUD_ENDPOINT, [SPACING, SPACING])[1:]

            # Build row of methods data
            if isinstance(endpoint.get(METHODS, NO_DATA), list) and endpoint.get(METHODS, []):
                methods_string = "; ".join(endpoint.get(METHODS, []))

            # Generate the list of endpoint data to be displayed. Each row displays an element in list,
            # where each element is a list of the columns.
            entry_list.append(
                [
                    endpoint.get("LogicalResourceId", NO_DATA),
                    endpoint.get("PhysicalResourceId", NO_DATA),
                    cloud_endpoint_furl_string,
                    methods_string,
                ]
            )

            # Add a spacing column with the next endpoint in the table in case there are multiple endpoints to display.
            if cloud_endpoint_furl_multi_list:
                for url in cloud_endpoint_furl_multi_list:
                    entry_list.append([SPACING, SPACING, url, SPACING])

        # Build out the table with the data collected to represent the endpoints
        table_data = {
            "format_string": "{Resource ID:<{0}} {Physical ID:<{1}} {Cloud Endpoints:<{2}} {Methods:<{3}}",
            "format_args": OrderedDict(
                {
                    "Resource ID": "Resource ID",
                    "Physical ID": "Physical ID",
                    "Cloud Endpoints": "Cloud Endpoints",
                    "Methods": "Methods",
                }
            ),
            "table_name": "Endpoints",
            "data": entry_list,
        }
        return table_data



""" Responsible for building schema code directory hierarchy based on schema name """

import re

# To sanitize schema package name and root name. During code generation schema service follows
# convention to replace all the special character except [a-zA-Z0-9_@] via _.
CHARACTER_TO_SANITIZE = "[^a-zA-Z0-9_@]"
POTENTIAL_PACKAGE_SEPARATOR = "[@]"


def get_package_hierarchy(schema_name):
    path = "schema"
    if schema_name.startswith("aws.partner-"):
        path = path + "." "aws.partner"
        tail = schema_name[len("aws.partner-") :]
        path = path + "." + sanitize_name(tail)
        return path.lower()
    if schema_name.startswith("aws."):
        parts = schema_name.split(".")
        for part in parts:
            path = path + "."
            path = path + sanitize_name(part)
        return path.lower()
    return f"{path}.{sanitize_name(schema_name)}".lower()


def sanitize_name(name):
    name = re.sub(CHARACTER_TO_SANITIZE, "_", name)
    return re.sub(POTENTIAL_PACKAGE_SEPARATOR, ".", name)



"""
Hook-related utils
"""

from typing import Dict, Optional

HOOK_METADATA_KEY = "AWS::SAM::Hook"


def get_hook_metadata(template: Dict[str, Dict]) -> Optional[Dict]:
    """
    Returns Hook Metadata from the given template if it exists.
    If the metadata is not found, returns None.

    Parameters
    ----------
    template:
        CFN template to look through

    Returns
    ----------
    Optional[Dict]:
        Hook metadata

    """

    metadata = template.get("Metadata", {})
    return metadata.get(HOOK_METADATA_KEY)



"""
Contains message used by Schemas paginated CLI.
"""


def construct_cli_display_message_for_schemas(page_to_render, last_page_number=None):
    if last_page_number is None:
        last_page_number = "many"
    single_page = "Event Schemas"
    first_page = "Event Schemas [Page %s/%s] (Enter N for next page)" % (page_to_render, last_page_number)
    middle_page = "Event Schemas [Page %s/%s] (Enter N/P for next/previous page)" % (page_to_render, last_page_number)
    last_page = "Event Schemas [Page %s/%s] (Enter P for previous page)" % (page_to_render, last_page_number)
    return {"single_page": single_page, "first_page": first_page, "middle_page": middle_page, "last_page": last_page}


def construct_cli_display_message_for_registries(page_to_render, last_page_number=None):
    if last_page_number is None:
        last_page_number = "many"
    single_page = "Schema Registry"
    first_page = "Schema Registry [Page %s/%s] (Enter N for next page)" % (page_to_render, last_page_number)
    middle_page = "Schema Registry [Page %s/%s] (Enter N/P for next/previous page)" % (page_to_render, last_page_number)
    last_page = "Schema Registry [Page %s/%s] (Enter P for previous page)" % (page_to_render, last_page_number)
    return {"single_page": single_page, "first_page": first_page, "middle_page": middle_page, "last_page": last_page}



"""Base SyncFlow for Lambda Function"""
import logging
import time
from abc import ABC
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List, cast

from botocore.client import BaseClient

from samcli.lib.providers.provider import Function, Stack
from samcli.lib.providers.sam_function_provider import SamFunctionProvider
from samcli.lib.sync.flows.alias_version_sync_flow import AliasVersionSyncFlow
from samcli.lib.sync.sync_flow import SyncFlow
from samcli.local.lambdafn.exceptions import FunctionNotFound

if TYPE_CHECKING:  # pragma: no cover
    from samcli.commands.build.build_context import BuildContext
    from samcli.commands.deploy.deploy_context import DeployContext
    from samcli.commands.sync.sync_context import SyncContext

LOG = logging.getLogger(__name__)
FUNCTION_SLEEP = 1  # used to wait for lambda function last update to be successful


class FunctionSyncFlow(SyncFlow, ABC):
    _function_identifier: str
    _function_provider: SamFunctionProvider
    _function: Function
    _lambda_client: Any
    _lambda_waiter: Any
    _lambda_waiter_config: Dict[str, Any]

    def __init__(
        self,
        function_identifier: str,
        build_context: "BuildContext",
        deploy_context: "DeployContext",
        sync_context: "SyncContext",
        physical_id_mapping: Dict[str, str],
        stacks: List[Stack],
    ):
        """
        Parameters
        ----------
        function_identifier : str
            Function resource identifier that need to be synced.
        build_context : BuildContext
            BuildContext
        deploy_context : DeployContext
            DeployContext
        sync_context: SyncContext
            SyncContext object that obtains sync information.
        physical_id_mapping : Dict[str, str]
            Physical ID Mapping
        stacks : Optional[List[Stack]]
            Stacks
        """
        super().__init__(
            build_context,
            deploy_context,
            sync_context,
            physical_id_mapping,
            log_name="Lambda Function " + function_identifier,
            stacks=stacks,
        )
        self._function_identifier = function_identifier
        self._function_provider = self._build_context.function_provider
        self._function = cast(Function, self._function_provider.get(self._function_identifier))
        self._lambda_client = None
        self._lambda_waiter = None
        self._lambda_waiter_config = {"Delay": 1, "MaxAttempts": 60}

    def set_up(self) -> None:
        super().set_up()
        self._lambda_client = self._boto_client("lambda")
        self._lambda_waiter = self._lambda_client.get_waiter("function_updated")

    @property
    def sync_state_identifier(self) -> str:
        """
        Sync state is the unique identifier for each sync flow
        In sync state toml file we will store
        Key as ZipFunctionSyncFlow:FunctionLogicalId
        Value as function ZIP hash
        """
        return self.__class__.__name__ + ":" + self._function_identifier

    def gather_dependencies(self) -> List[SyncFlow]:
        """Gathers alias and versions related to a function.
        Currently only handles serverless function AutoPublishAlias field
        since a manually created function version resource behaves statically in a stack.
        Redeploying a version resource through CFN will not create a new version.
        """
        LOG.debug("%sWaiting on Remote Function Update", self.log_prefix)
        self._lambda_waiter.wait(
            FunctionName=self.get_physical_id(self._function_identifier), WaiterConfig=self._lambda_waiter_config
        )
        LOG.debug("%sRemote Function Updated", self.log_prefix)
        sync_flows: List[SyncFlow] = list()

        function_resource = self._get_resource(self._function_identifier)
        if not function_resource:
            raise FunctionNotFound(f"Unable to find function {self._function_identifier}")

        auto_publish_alias_name = function_resource.get("Properties", dict()).get("AutoPublishAlias", None)
        if auto_publish_alias_name:
            sync_flows.append(
                AliasVersionSyncFlow(
                    self._function_identifier,
                    auto_publish_alias_name,
                    self._build_context,
                    self._deploy_context,
                    self._sync_context,
                    self._physical_id_mapping,
                    self._stacks,
                )
            )
            LOG.debug("%sCreated Alias and Version SyncFlow", self.log_prefix)

        return sync_flows

    def _equality_keys(self):
        return self._function_identifier


class FunctionUpdateStatus(Enum):
    """Function update return types"""

    SUCCESS = "Successful"
    FAILED = "Failed"
    IN_PROGRESS = "InProgress"


def wait_for_function_update_complete(lambda_client: BaseClient, physical_id: str) -> None:
    """
    Checks on cloud side to wait for the function update status to be complete

    Parameters
    ----------
    lambda_client : boto.core.BaseClient
        Lambda client that performs get_function API call.
    physical_id : str
        Physical identifier of the function resource
    """

    status = FunctionUpdateStatus.IN_PROGRESS.value
    while status == FunctionUpdateStatus.IN_PROGRESS.value:
        response = lambda_client.get_function(FunctionName=physical_id)  # type: ignore
        status = response.get("Configuration", {}).get("LastUpdateStatus", "")

        if status == FunctionUpdateStatus.IN_PROGRESS.value:
            time.sleep(FUNCTION_SLEEP)

    LOG.debug("Function update status on %s is now %s on cloud.", physical_id, status)



"""Hook Package Config"""
import json
from pathlib import Path
from typing import Dict, NamedTuple, Optional, cast

import jsonschema  # type: ignore

from .exceptions import InvalidHookPackageConfigException


class HookFunctionality(NamedTuple):
    """
    A class to represent a hook functionality (e.g. prepare)
    """

    entry_method: Dict[str, str]

    @property
    def module(self) -> str:
        return self.entry_method["module"]

    @property
    def method(self) -> str:
        return self.entry_method["method"]


class HookPackageConfig:
    """
    A class to represent a hook package. Upon instantiation, it also validate the config against a json schema.
    """

    _package_dir: Path
    _config: Dict

    CONFIG_FILENAME = "Config.json"
    JSON_SCHEMA_PATH = Path(__file__).parent / "hook_config_schema.json"

    def __init__(self, package_dir: Path):
        """
        Parameters
        ----------
        package_dir: Path
            The path of the hook package directory
        """
        self._package_dir = package_dir
        config_loc = package_dir / self.CONFIG_FILENAME
        if not config_loc.is_file():
            raise InvalidHookPackageConfigException(f"{config_loc} is not a file or does not exist")

        with config_loc.open("r", encoding="utf-8") as f:
            config_dict = json.load(f)

        try:
            jsonschema.validate(config_dict, self.jsonschema)
        except jsonschema.ValidationError as e:
            raise InvalidHookPackageConfigException(f"Invalid Config.json - {e}") from e

        for func, func_dict in config_dict["functionalities"].items():
            config_dict["functionalities"][func] = HookFunctionality(func_dict["entry_method"])
        self._config = config_dict

    @property
    def jsonschema(self) -> Dict:
        with HookPackageConfig.JSON_SCHEMA_PATH.open("r", encoding="utf-8") as f:
            jsonschema_dict = json.load(f)
        return cast(Dict, jsonschema_dict)

    @property
    def name(self) -> str:
        return cast(str, self._config["hook_name"])

    @property
    def use_case(self) -> str:
        return cast(str, self._config["hook_use_case"])

    @property
    def version(self) -> str:
        return cast(str, self._config["version"])

    @property
    def specification(self) -> str:
        return cast(str, self._config["hook_specification"])

    @property
    def description(self) -> Optional[str]:
        return cast(str, self._config.get("description"))

    @property
    def functionalities(self) -> Dict[str, HookFunctionality]:
        return cast(Dict[str, HookFunctionality], self._config["functionalities"])

    @property
    def iac_framework(self) -> str:
        return cast(str, self._config.get("iac_framework", ""))
